"""Testing new plugin+resource selector+plugin runner components
"""

import os
import sys
sys.path.append('./src/')
import json

from fusion_tools.handler.dsa_handler import DSAHandler
from fusion_tools.handler.dsa_handler import DSAPluginRunner

from dash import dcc
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
from dash_extensions.enrich import DashProxy, MultiplexerTransform, html

test_plugin = {
    'name': 'NucleiDetection',
    'image': 'dsarchive/histomicstk:latest',
    'input_args': [
        {
            'name': 'inputImageFile',
            'default': {
                'type': 'input_file',
                'name': 'Image'
            },
            'disabled': True
        },
        {
            'name': 'outputNucleiAnnotationFile_folder',
            'default': {
                'type': 'upload_folder'
            },
            'disabled': True
        },
        {
            'name': 'outputNucleiAnnotationFile',
            'default': {
                'value': 'nuclei_annotations.annot'
            },
            'disabled': True
        },
        'nuclei_annotation_format',
        'min_nucleus_area',
        'ignore_border_nuclei',
        'ImageInversionForm'
    ]
}

def main():
    
    base_url = 'http://ec2-3-230-122-132.compute-1.amazonaws.com:8080/api/v1'

    user_name = os.getenv('DSA_USER')
    p_word = os.getenv('DSA_PWORD')

    # You have to sign in to access the add_plugin() method
    dsa_handler = DSAHandler(
        girderApiUrl=base_url,
        username = user_name,
        password = p_word
    )

    plugin_runner_component = DSAPluginRunner(
        handler = dsa_handler
    )

    session_data = {
        'current_user': dsa_handler.authenticate_new(
            username = user_name,
            password = p_word
        )
    }

    plugin_runner_component.load(0)
    plugin_runner_component.gen_layout(session_data)

    main_app = DashProxy(
        __name__,
        external_stylesheets = [
            dbc.themes.LUX,
            dbc.themes.BOOTSTRAP,
            dbc.icons.BOOTSTRAP,
            dbc.icons.FONT_AWESOME,
            dmc.styles.ALL,
        ],
        transforms = [
            MultiplexerTransform()
        ]
    )
    main_app.layout = dmc.MantineProvider(html.Div(
        [
            dcc.Store(
                id = 'anchor-vis-store',
                data = json.dumps(session_data),
                storage_type='memory'
            ),
            dbc.Card([
                dbc.CardHeader(plugin_runner_component.title),
                dbc.CardBody([
                    plugin_runner_component.blueprint.embed(main_app),
                ])
            ])
        ]
    ))

    main_app.run(
        port = '8050',
        debug=True
    )

if __name__=='__main__':
    main()





"""Testing FUSION layout
"""

import os
import sys
sys.path.append('./src/')

from fusion_tools.fusion.vis import get_layout

def main():
    
    dsa_url = os.environ.get('DSA_URL')
    dsa_user = os.environ.get('DSA_USER')
    dsa_pword = os.environ.get('DSA_PWORD')

    if all([i is None for i in [dsa_url,dsa_user,dsa_pword]]):
        raise Exception('Need to initialize with at least the environment variable: DSA_URL')

    initial_items = [
        '6495a4e03e6ae3107da10dc5',
        '6495a4df3e6ae3107da10dc2'
    ] 

    args_dict = {
        'girderApiUrl': dsa_url,
        'user': dsa_user,
        'pword': dsa_pword,
        'initialItems': initial_items,
        'app_options': {
            'port': 8050
        }
    }

    fusion_vis = get_layout(args_dict)

    fusion_vis.start()

if __name__=='__main__':
    main()



"""
Testing new Visualization structure

"""

import os
import sys
sys.path.append('./src/')

from fusion_tools import Visualization
from fusion_tools.components import SlideMap, OverlayOptions, PropertyViewer, PropertyPlotter
from fusion_tools.handler.dsa_handler import DSAHandler


def main():
 
    # Test 1) Visualization with slides and annotations as input arguments that is then 
    # extended to all components where that is required  
    # Test 2) TileServers with multiple slides, switching between slides and 
    # their respective annotations. (DSA and Local)
    # Test 3) Reconfigure "start" function to allow for expanded app_options and 
    # servers. Try and bundle LocalTileServer.start() in here as well to run silently/verbosely

    # There should be some option for just creating a server which has images and their annotations
    # and then referencing them with something similar to the DSAHandler (LocalDataHandler?).
    # Should allow for a DSAHandler class with some list of names/item Ids.

    # Main thing is that there should be a way to establish that one set of annotations belongs to one slide

    base_dir = 'C:\\Users\\samuelborder\\Desktop\\HIVE_Stuff\\FUSION\\Test Upload\\'
    local_slide_list = [
        base_dir+'XY01_IU-21-015F_001.svs',
        base_dir+'XY01_IU-21-015F.svs',
        base_dir+'new Visium\\V12U21-010_XY02_21-0069.tif',
    ]
    local_annotations_list = [
        base_dir+'XY01_IU-21-015F_001.xml',
        None,
        base_dir+'new Visium\\V12U21-010_XY02_21-0069.h5ad',
    ]

    dsa_handler = DSAHandler(
        girderApiUrl = 'http://ec2-3-230-122-132.compute-1.amazonaws.com:8080/api/v1'
    )

    dsa_items_list = [
        '64f545082d82d04be3e39ee1',
        '64f54be52d82d04be3e39f65'
    ]

    dsa_tileservers = [dsa_handler.get_tile_server(i) for i in dsa_items_list]
    
    vis_sess = Visualization(
        local_slides = local_slide_list,
        local_annotations = local_annotations_list,
        tileservers = dsa_tileservers,
        linkage = 'col',
        components = [
            [
                [
                    SlideMap(),
                    OverlayOptions(ignore_list = ['_id','_index','barcode']),
                    PropertyViewer(ignore_list = ['_id','_index','barcode']),
                    PropertyPlotter(ignore_list = ['_id','_index','barcode'])
                ],            
                [
                    SlideMap(),
                    OverlayOptions(ignore_list = ['_id','_index','barcode']),
                    PropertyViewer(ignore_list = ['_id','_index','barcode']),
                    PropertyPlotter(ignore_list = ['_id','_index','barcode'])
                ]
            ]
        ],
        app_options={'port': 8050}
    )

    vis_sess.start()



if __name__=='__main__':
    main()




"""Testing plugin conductor and plugins
"""

import os
import sys
sys.path.append('./src/')

from fusion_tools.handler.dsa_handler import DSAHandler
from girder_job_sequence.utils import from_list

import requests
import json



def main(args):
    print(args)

    base_url = 'http://ec2-3-230-122-132.compute-1.amazonaws.com:8080/api/v1'

    user_name = os.getenv('DSA_USER')
    p_word = os.getenv('DSA_PWORD')

    # You have to sign in to access the add_plugin() method
    dsa_handler = DSAHandler(
        girderApiUrl=base_url,
        username = user_name,
        password = p_word
    )
    
    if args[1]=='add':
        if len(args)>2:
            plugin_add_list = args[2:]
        else:
            plugin_add_list = [
                'fusionplugins/codex:latest',
                'fusionplugins/general:latest',
                'fusionplugins/visium:latest'
            ]
        
        # This adds the plugins in plugin_add_list, removing previous versions if they are present
        response = dsa_handler.add_plugin(
            image_name = plugin_add_list
        )

        print(response)

    elif args[1]=='run':

        if args[2]=='multifeat':
            # MultiCompartment segmentation --> FeatureExtraction
            job_list = [
                {
                    'docker_image': 'samborder2256/multicomp:latest',
                    'cli': 'MultiCompartmentSegment',
                    'input_args': [
                        {
                            'name': 'files',
                            'value': '6717e743f433060d2884838c'
                        },
                        {
                            'name': 'base_dir',
                            'value': '6717e73cf433060d28848389'
                        },
                        {
                            'name': 'modelfile',
                            'value': '648123761019450486d13dce'
                        }
                    ]
                },
                {
                    'docker_image': 'fusionplugins/general:latest',
                    'cli': 'FeatureExtraction',
                    'input_args': [
                        {
                            'name': 'input_image',
                            'value': "{{'type':'file','item_type':'path','item_query':'/user/sam123/Public/FUSION_Upload_2024_10_22_13_56_05_219929/XY01_IU-21-015F_001.svs','file_type':'fileName','file_query':'XY01_IU-21-015F_001.svs'}}"
                        },
                        {
                            'name': 'extract_sub_compartments',
                            'value': True
                        },
                        {
                            'name': 'hematoxylin_threshold',
                            'value': 150
                        },
                        {
                            'name': 'eosinophilic_threshold',
                            'value': 30
                        },
                        {
                            'name': 'hematoxylin_min_size',
                            'value': 40
                        },
                        {
                            'name': 'eosinophilic_min_size',
                            'value': 20
                        },
                    ]
                }
            ]

        elif args[2]=='codexfeat':
            # Cell segmentation --> FeatureExtraction
            job_list = [
                {
                    'docker_image': 'fusionplugins/codex:latest',
                    'cli': 'CellSegmentation',
                    'input_args': [
                        {
                            'name': 'input_image',
                            'value': '67af56bafcdeba1e293215e4'
                        },
                        {
                            'name': 'input_region',
                            'value': "[1600,1000,2600,2000]"
                        },
                        {
                            'name': 'return_segmentation_region',
                            'value': True
                        }
                    ]
                },
                {
                    'docker_image': 'fusionplugins/general:latest',
                    'cli': 'FeatureExtraction',
                    'input_args': [
                        {
                            'name': 'input_image',
                            'value': '67af56bafcdeba1e293215e4'
                        },
                        {
                            'name': 'extract_sub_compartments',
                            'value': False
                        }
                    ]
                }
            ]

        elif args[2]=='visium':
            # Cell deconvolution --> spot annotation --> feature extraction
            job_list = [
                {
                    'docker_image': 'fusionplugins/visium:latest',
                    'cli': 'CellDeconvolution',
                    'input_args': [
                        {
                            'name': 'counts_file',
                            'value': '67af4e59fcdeba1e293211c2'
                        },
                        {
                            'name': 'organ',
                            'value': 'Azimuth Kidney Reference'
                        }
                    ]
                },
                {
                    'docker_image': 'fusionplugins/visium:latest',
                    'cli': 'SpotAnnotation',
                    'input_args': [
                        {
                            'name': 'counts_file',
                            'value': "{{'type': 'file', 'item_type': '_id','item_query': '67af4e59fcdeba1e293211c0', 'file_type': 'fileName','file_query': 'V10S15-102_XY02_IU-21-019-5_integrated.rds'}}"
                        },
                        {
                            'name': 'input_files',
                            'value': '67af4e59fcdeba1e293211c0'
                        },
                        {
                            'name': 'use_gene_selection',
                            'value': True
                        },
                        {
                            'name': 'gene_selection_method',
                            'value': 'dispersion'
                        },
                        {
                            'name': 'n',
                            'value': 10
                        }
                    ]
                }
            ]
        
        job_sequence = from_list(dsa_handler.gc, job_list)
        for j in job_sequence.jobs:
            print(json.dumps(j.input_args,indent=4))
            #print(json.dumps(j.executable_dict,indent=4))

        job_sequence.start(verbose=True,cancel_on_error=True)
        
    

if __name__=='__main__':
    main(sys.argv)



# %% [markdown]
# ### Making annotations for *10x Visium* data
# 
# #### Step 1: (*external*) Convert *.RDS* file (SpaceRanger output) to *.h5ad*
# - This can be done using Seurat and SeuratDisk to go from *SeuratObject* -> *.h5Seurat* -> *.h5ad* see [this page](https://mojaveazure.github.io/seurat-disk/articles/convert-anndata.html) although the spot coordinates have to be added to the `seuratObject` manually in order to be added to the *.h5ad* file.
#     - Do this using:
#     
#         `seuratObject@meta.data[["spot_coordinates"]] <- cbind(seuratObject@meta.data, seuratObject@images[["slice1"]]@coordinates)`
# 
#     - **Important**:
#         - `Convert` will only save the *DefaultAssay* to the *X* matrix and will ignore all other assays. You can get around this by resetting the *DefaultAssay*:
# 
#             `DefaultAssay(seuratObject) <- "predsubclassl2"`
#     

# %% [markdown]
# #### Step 2: Create annotation files using *fusion-tools*

# %%
#%pip install fusion-tools[interactive]==2.3.3

# %%
import os
import sys
import json
sys.path.append('./src/')

import numpy as np
import pandas as pd

import anndata as ad
from tqdm import tqdm
import uuid

from shapely.geometry import Point, shape, Polygon
from shapely.ops import unary_union

from fusion_tools.utils.shapes import load_visium, spatially_aggregate
from fusion_tools.utils.omics import group_subtypes, selective_aggregation

visium_path = "..\\..\\Test Upload\\new Visium\\V12U21-010_XY02_21-0069.h5ad"

# This function will be in the next released version
spot_annotations = load_visium(visium_path)
anndata_object = ad.read_h5ad(visium_path)


# %% [markdown]
# ## Creating Cell Hierarchical labels

# %%
cell_groups_path = "..\\..\\Test Upload\\new Visium\\Supplementary Tables Version 1 05-2024.xlsx"
cell_groups = pd.read_excel(cell_groups_path,engine='openpyxl',sheet_name='Table S - sn annot',skiprows=range(4),index_col=0)
cell_groups = cell_groups[cell_groups["v2.subclass.l1"].notna()]

l1_types = cell_groups["v2.subclass.l1"].dropna(how='all').unique()
main_cell_grouping = {}
for l1 in l1_types:
    l2_types = cell_groups[cell_groups["v2.subclass.l1"].str.contains(l1)]["v2.subclass.l2"].tolist()
    main_cell_grouping[l1] = l2_types

print(json.dumps(main_cell_grouping,indent=4))


# %% [markdown]
# ## Adding Seurat Clusters as a sub-property for each spot

# %%
# Aligning annotations with cell subtype values and Seurat clusters (stored in obs):
pred_l2 = anndata_object.to_df().round(decimals = 5)
clusters = anndata_object.obs['seurat_clusters'].values
for row in range(pred_l2.shape[0]):
    row_data = pred_l2.iloc[row,:].to_dict()
    del row_data['max']

    spot_annotations['features'][row]['properties'] = spot_annotations['features'][row]['properties'] | {'L2 Cell Types': row_data, 'Seurat Cluster': str(clusters[row])}
    spot_annotations['features'][row]['properties'] = spot_annotations['features'][row]['properties'] | {"L1 Cell Types": group_subtypes({"L2": row_data},"L2",main_cell_grouping,keep_zeros=True)}
       


# %% [markdown]
# ## Extracting simple morphometrics for spots

# %%
from PIL import Image
from io import BytesIO
import large_image
from skimage.filters import threshold_otsu
from skimage.segmentation import watershed
from skimage import exposure
from skimage.color import rgb2hsv
import scipy.ndimage as ndi
from skimage.morphology import remove_small_holes, remove_small_objects
from skimage.measure import label, find_contours
from skimage.feature import peak_local_max
import matplotlib.pyplot as plt

from fusion_tools.feature_extraction import (
    ParallelFeatureExtractor, distance_transform_features, color_features,
    texture_features, morphological_features
)

if False:
    # define sub-compartment segmentation function
    def stain_mask(image,mask):

        seg_params = [
            {
                'name': 'Nuclei',
                'threshold': 150,
                'min_size': 40
            },
            {
                'name': 'Eosinophilic',
                'threshold': 30,
                'min_size': 20
            },
            {
                'name': 'Luminal Space',
                'threshold': 0,
                'min_size': 0
            }
        ]

        image_shape = np.shape(image)

        sub_comp_image = np.zeros((image_shape[0],image_shape[1],3))
        remainder_mask = np.ones((image_shape[0],image_shape[1]))

        hsv_image = np.uint8(255*rgb2hsv(image))
        hsv_image = hsv_image[:,:,1]

        for idx,param in enumerate(seg_params):

            # Check for if the current sub-compartment is nuclei
            if param['name'].lower()=='nuclei':
                # Using the inverse of the value channel for nuclei
                h_image = 255-np.uint8(255*rgb2hsv(image)[:,:,2])
                h_image = np.uint8(255*exposure.equalize_hist(h_image))

                remaining_pixels = np.multiply(h_image,remainder_mask)
                masked_remaining_pixels = np.multiply(remaining_pixels,mask)
                #masked_remaining_pixels = remaining_pixels

                # Applying manual threshold
                masked_remaining_pixels[masked_remaining_pixels<=param['threshold']] = 0
                masked_remaining_pixels[masked_remaining_pixels>0] = 1

                # Area threshold for holes is controllable for this
                sub_mask = remove_small_holes(masked_remaining_pixels>0,area_threshold=10)
                sub_mask = sub_mask>0
                # Watershed implementation from: https://scikit-image.org/docs/stable/auto_examples/segmentation/plot_watershed.html
                distance = ndi.distance_transform_edt(sub_mask)
                labeled_mask, _ = ndi.label(sub_mask)
                coords = peak_local_max(distance,footprint=np.ones((3,3)),labels = labeled_mask)
                watershed_mask = np.zeros(distance.shape,dtype=bool)
                watershed_mask[tuple(coords.T)] = True
                markers, _ = ndi.label(watershed_mask)
                sub_mask = watershed(-distance,markers,mask=sub_mask)
                sub_mask = sub_mask>0

                # Filtering out small objects again
                sub_mask = remove_small_objects(sub_mask,param['min_size'])

            else:

                remaining_pixels = np.multiply(hsv_image,remainder_mask)
                masked_remaining_pixels = np.multiply(remaining_pixels,mask)
                #masked_remaining_pixels = remaining_pixels

                # Applying manual threshold
                masked_remaining_pixels[masked_remaining_pixels<=param['threshold']] = 0
                masked_remaining_pixels[masked_remaining_pixels>0] = 1

                # Filtering by minimum size
                small_object_filtered = (1/255)*np.uint8(remove_small_objects(masked_remaining_pixels>0,param['min_size']))

                sub_mask = small_object_filtered

            sub_comp_image[sub_mask>0,idx] = 1
            remainder_mask -= sub_mask>0

        # Assigning remaining pixels within the boundary mask to the last sub-compartment
        #remaining_pixels = np.multiply(mask,remainder_mask)
        remaining_pixels = remainder_mask
        sub_comp_image[remaining_pixels>0,idx] = 1

        final_mask = np.zeros_like(remainder_mask)
        final_mask += sub_comp_image[:,:,0]
        final_mask += 2*sub_comp_image[:,:,1]
        final_mask += 3*sub_comp_image[:,:,2]

        return final_mask


    # Define feature extractor class
    feature_extractor = ParallelFeatureExtractor(
        image_source = "..\\..\\Test Upload\\new Visium\\V12U21-010_XY02_21-0069.tif",
        feature_list = [
                lambda image,mask,coords: distance_transform_features(image,mask,coords),
                lambda image,mask,coords: color_features(image,mask,coords),
                lambda image,mask,coords: texture_features(image,mask,coords),
                lambda image,mask,coords: morphological_features(image,mask,coords)
        ],
        preprocess = None,
        sub_mask = lambda image,mask: stain_mask(image,mask),
        mask_names = ['Nuclei','Eosinophilic','Luminal Space'],
        channel_names = ['Red','Green','Blue'],
        n_jobs = 4,
        verbose = True
    )
    # Execute feature extraction
    feature_df = feature_extractor.start(spot_annotations['features'])




    # %%
    # Add properties to spot annotations
    spot_features = feature_df.to_dict('records')
    for f,s in zip(spot_features,spot_annotations['features']):
        s['properties'] = s['properties'] | {"Pathomics": f}

    print(json.dumps(spot_annotations['features'][0]['properties'],indent=4))

# %% [markdown]
# ## Merging adjacent spots from the same cluster

# %%
# Merging adjacent spots based on Seurat cluster
poly_list = []
for f in spot_annotations['features']:
    poly_list.append(Point(shape(f['geometry']).centroid))

# Making the tissue mask
def find_tissue(slide_tile_source):

    # Grabbing the thumbnail of the image (RGB)
    slide_metadata = slide_tile_source.getMetadata()
    thumbnail_img,_ = slide_tile_source.getThumbnail()
    thumb_array = np.array(Image.open(BytesIO(thumbnail_img)))

    # Getting scale factors for thumbnail image to full-size image
    thumbX, thumbY = np.shape(thumb_array)[1],np.shape(thumb_array)[0]
    scale_x = slide_metadata['sizeX']/thumbX
    scale_y = slide_metadata['sizeY']/thumbY

    # Mean of all channels/frames to make grayscale mask
    gray_mask = np.squeeze(np.mean(thumb_array,axis=-1))

    threshold_val = 1.15*threshold_otsu(gray_mask)
    tissue_mask = gray_mask <= threshold_val

    tissue_mask = remove_small_holes(tissue_mask,area_threshold=150)
    tissue_mask = remove_small_objects(tissue_mask,min_size=64)

    #plt.imshow(tissue_mask)
    #plt.show()
    labeled_mask = label(tissue_mask)
    tissue_pieces = np.unique(labeled_mask).tolist()
    tissue_areas = [np.sum(labeled_mask==j) for j in tissue_pieces[1:]]
    largest_2 = [tissue_pieces[1:][tissue_areas.index(m)] for m in tissue_areas if m in sorted(tissue_areas)[-2:]]

    tissue_shape_list = []
    for piece in largest_2:
        if np.sum(labeled_mask==piece)>100:
            tissue_contours = find_contours(labeled_mask==piece)
            for contour in tissue_contours:
                poly_list = [(i[1]*scale_x,i[0]*scale_y) for i in contour]
                if len(poly_list)>2:
                    obj_polygon = Polygon(poly_list)
                    if obj_polygon.is_valid:
                        tissue_shape_list.append(obj_polygon)

    # Merging shapes together to remove holes
    merged_tissue = unary_union(tissue_shape_list)
    if merged_tissue.geom_type=='Polygon':
        merged_tissue = [merged_tissue]
    elif merged_tissue.geom_type in ['MultiPolygon','GeometryCollection']:
        merged_tissue = merged_tissue.geoms

    thumbnail_geojson = {
        'type': 'FeatureCollection',
        'features': [
            {
                'type':'Feature',
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': [list(i.exterior.coords)]
                },
                'properties': {
                    'name': 'Tissue Mask',
                    '_id': uuid.uuid4().hex[:24],
                    '_index': idx
                }
            }
            for idx,i in enumerate(merged_tissue) if i.geom_type=='Polygon'
        ],
        'properties': {
            'name': 'Tissue Mask',
            '_id': uuid.uuid4().hex[:24]
        }
    }

    return thumbnail_geojson


tissue_mask = find_tissue(
    large_image.open('..\\..\\Test Upload\\new Visium\\V12U21-010_XY02_21-0069.tif')
)

unique_clusters = np.unique(clusters).tolist()
merged_geojson = []
for c in unique_clusters:
    c_idx = np.where(clusters==c)[0].tolist()

    cluster_spots = [poly_list[i] for i in c_idx]
    cluster_merges = {
        'type': 'FeatureCollection',
        'features': [],
        'properties': {
            'name': f'Cluster {c}',
            '_id': uuid.uuid4().hex[:24]
        }
    }
    for t in tissue_mask['features']:
        tissue_shape = shape(t['geometry'])
        within_tissue_spots = [i for i in cluster_spots if i.within(tissue_shape)]
        if len(within_tissue_spots)>0:
            # Buffering spots to a size that they start to intersect
            # (for Visium spots this is a little easier to define since all of the spots are separated 
            # by the same distance, for other structures this might have to be some kind of Voronoi diagram)
            buffered_spots = [i.buffer(100) for i in within_tissue_spots]

            # Merging buffered spots
            merged_spots = unary_union(buffered_spots)
            cluster_geojson = [
                {
                    'type': 'Feature',
                    'geometry': {
                        'type': 'Polygon',
                        'coordinates': [list(g.exterior.coords)]
                    },
                    'properties': {
                        'name': f'Cluster {c}',
                        '_id': uuid.uuid4().hex[:24],
                        '_index': g_idx
                    }
                }
                for g_idx,g in enumerate(merged_spots.geoms)
            ]

            cluster_merges['features'].extend(cluster_geojson)
            
    merged_geojson.append(cluster_merges)

spatially_agged = []
for m in merged_geojson:
    spatially_agged.append(spatially_aggregate(m,[spot_annotations]))


# %% [markdown]
# ## Visualization in *fusion-tools*

# %%
# Visualizing spots in fusion-tools
from fusion_tools.components import SlideMap,OverlayOptions,PropertyPlotter,HRAViewer, BulkLabels
from fusion_tools import Visualization

local_image_path = "..\\..\\Test Upload\\new Visium\\V12U21-010_XY02_21-0069.tif"

vis_session = Visualization(
    local_slides = [local_image_path],
    local_annotations = [spot_annotations],
    components = [
        [
            SlideMap(),
            [
                OverlayOptions(),
                PropertyPlotter(),
                BulkLabels(),
                HRAViewer()
            ]
        ]
    ],
    app_options = {
        'jupyter': False,
        'port': 8080
    }
)

vis_session.start()






"""test recursive property extraction
"""
import json

def read_dict_list(dict_list,main_list = [],sub_list = None):
    for d in dict_list:
        if type(d)==dict:
            title = list(d.keys())[0]
            print(f'title: {title}')

            nested = [{key:val} for key,val in d[title].items() if type(val)==dict]
            non_nested = [{'SubProperty': key, 'Value': val} for key,val in d[title].items() if not type(val)==dict]
            main_list.append(non_nested)
            read_dict_list(nested,main_list,[])
    return main_list


def main():

    test_list = [
        {
            'main_prop': {
                'sub_prop1': {},
                'sub_prop2': {
                    'sub_subprop1': 2
                },
                'sub_prop3': {
                    'sub_subprop1': {
                        'sub_key': 'sub_val',
                        'sub_key2': 'sub_val2'
                    },
                    'sub_subprop2': {
                        'sub_sub_subprop1': 4
                    }
                }
            }
        },
        {
            'main_prop2': {
                'sub_prop2': {
                    'key1': 'val1',
                    'key2': 'val2',
                    'key3': 'val3',
                    'nested_key': {
                        'n_key1': 'n_val1',
                        'n_key2': 'n_val2'
                    }
                }
            }
        }
    ]

    print(json.dumps(test_list,indent=4))

    print(read_dict_list(test_list))



if __name__=='__main__':
    main()





"""Testing spatial aggregation function (different versions)
"""

import os
import sys
sys.path.append('./src/')
import uuid

from fusion_tools.utils.shapes import spatially_aggregate
from fusion_tools.handler.dsa_handler import DSAHandler

def main():

    # Get annotations (with some overlapping and some not overlapping)
    # Add properties to annotations (multiple different nested levels)
    # Test 1) spatial aggregation, separate=True, summarize=True
    # Test 2) spatial aggregation, separate=False, summarize=True
    # Test 3) spatial aggregation, separate = True, summarize = False
    # Test 4) spatial aggregation, separate = False, summarize = False

    dsa_handler = DSAHandler(
        girderApiUrl='http://ec2-3-230-122-132.compute-1.amazonaws.com:8080/api/v1'
    )

    annotations = dsa_handler.get_annotations(
        item = '64f545302d82d04be3e39eec'
    )

    print(f'len of annotations: {len(annotations)}')
    print([i['properties']['name'] for i in annotations])

    # Clearing properties from the glomeruli:
    glomeruli = annotations[0]
    spots = annotations[1]

    for f_idx, f in enumerate(glomeruli['features']):
        f['properties'] = {
            'name': 'Glomeruli',
            '_id': uuid.uuid4().hex[:24],
            '_index': f_idx
        }
    
    print(glomeruli['features'][0]['properties'])

    ## Test 1: spatial aggregation, separate = True, summary = True
    agg_gloms1 = spatially_aggregate(glomeruli,[spots],separate=True,summarize=True)
    print('--------------------separate = True, summarize = True---------------------------')
    print(agg_gloms1['features'][0]['properties'])
    print('\n\n\n')

    ## Test 2: spatial aggregation, separate = False, summary = True
    agg_gloms2 = spatially_aggregate(glomeruli,[spots],separate=False,summarize=True)
    print('--------------------separate = False, summarize = True---------------------------')
    print(agg_gloms2['features'][0]['properties'])
    print('\n\n\n')

    ## Test 3: spatial aggregation, separate = True, summary = False
    agg_gloms3 = spatially_aggregate(glomeruli,[spots],separate=True,summarize=False)
    print('--------------------separate = True, summarize = False---------------------------')
    print(agg_gloms3['features'][0]['properties'])
    print('\n\n\n')
    
    ## Test 4: spatial aggregation, sepate = False, summary = False
    agg_gloms4 = spatially_aggregate(glomeruli,[spots],separate=False,summarize=False)
    print('--------------------separate = False, summarize = False---------------------------')
    print(agg_gloms4['features'][0]['properties'])
    print('\n\n\n')





if __name__=='__main__':
    main()




# %%

# Importing packages
import os
import sys
import json
sys.path.append('../src/')
#!{sys.executable} -m pip install segmentation-models-pytorch plotly kaleido==0.1.0 albumentations
#!{sys.executable} -m pip install nbformat
import numpy as np
from tqdm import tqdm
import pandas as pd

import torch
import torchvision
import segmentation_models_pytorch as smp
from torch.utils.data import DataLoader
from torchvision.transforms import ToTensor, Normalize
from torchvision.transforms.v2 import RandomHorizontalFlip, RandomVerticalFlip, ElasticTransform, ColorJitter, Compose
import albumentations as A

from typing import Callable, List
from fusion_tools.dataset import ClassificationDataset
from fusion_tools.utils.shapes import load_histomics
from math import floor

import plotly.express as px
import plotly.graph_objects as go

# %%
class CellClassificationModel(torch.nn.Module):
    def __init__(self,
                 output_size: int = 2,
                 simple: bool = True):
        super().__init__()
        self.output_size = output_size
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.simple = simple
        self.fc1 = torch.nn.LazyLinear(120)
        self.fc2 = torch.nn.Linear(120,84)
        self.fc3 = torch.nn.Linear(84,self.output_size)
        
        if self.simple:
            self.conv1 = torch.nn.Conv2d(3,6,5)
            self.pool = torch.nn.MaxPool2d(2,2)
            self.conv2 = torch.nn.Conv2d(6,16,5)

            self.conv_layers = torch.nn.Sequential(
                self.conv1,
                torch.nn.ReLU(inplace=True),
                self.pool,
                self.conv2,
                torch.nn.ReLU(inplace=True),
                self.pool
            )

        else:
            model_weights = torchvision.models.EfficientNet_V2_S_Weights
            self.conv_layers = torchvision.models.efficientnet_v2_s(weights = model_weights.DEFAULT)

        self.linear_layers = torch.nn.Sequential(
            self.fc1,
            torch.nn.ReLU(inplace=True),
            self.fc2,
            torch.nn.ReLU(inplace=True),
            self.fc3
        )

    def forward(self, input):
        
        output = self.conv_layers(input)
        output = torch.flatten(output,1)
        output = torch.nn.Sigmoid()(self.linear_layers(output))        
        return output


def cell_percentages(inp:list):
    if len(inp)>0:
        return torch.from_numpy(np.array([sum([-1*(i-1) for i in inp])/len(inp), sum(inp)/len(inp)]))
    else:
        return torch.from_numpy(np.array([1.0, 0.0]))

def train(train_data, val_data, model, optimizer, loss, output_dir):
    
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    model.to(device)
    loss.to(device)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    epoch_num = 5
    train_loss = 0
    val_loss = 0
    losses = []
    train_step_count = 0
    with tqdm(total = epoch_num, position = 0, leave = True, file = sys.stdout) as pbar:
        for i in range(0,epoch_num):
            for step, (train_imgs,train_labels) in enumerate(train_data):
                model.train()
                optimizer.zero_grad()

                pbar.set_description(f'Epoch: {i}/{epoch_num}, Train Step: {step}, Train/Val Loss: {round(train_loss,4)}/{round(val_loss,4)}')

                train_imgs = train_imgs.to(device)
                train_labels = train_labels.to(device)

                train_pred = model(train_imgs)

                train_loss = loss(train_pred, train_labels)
                train_loss.backward()
                train_loss = train_loss.item()

                optimizer.step()
                losses.append({
                    'step': train_step_count+step, 'loss': train_loss, 'Train/Val': 'train'
                })

                # Performing validation
                with torch.no_grad():
                    model.eval()
                    val_imgs, val_labels = next(iter(val_data))

                    val_imgs = val_imgs.to(device)
                    val_labels = val_labels.to(device)

                    val_pred = model(val_imgs)
                    val_loss = loss(val_pred, val_labels)
                    val_loss = val_loss.item()

                    losses.append(
                        {'step': train_step_count+step, 'loss': val_loss, 'Train/Val': 'val'}
                    )

            train_step_count+=step
            torch.save(model.state_dict(), output_dir+'Classification_Model.pth')
            loss_df = pd.DataFrame.from_records(losses)
            loss_df.to_csv(output_dir+'Loss.csv')
            vis_loss(loss_df,output_dir)

            pbar.update(1)


        pbar.update(1)
        torch.save(model.state_dict(), output_dir+'Classification_Model.pth')
        loss_df = pd.DataFrame.from_records(losses)
        loss_df.to_csv(output_dir+'Loss.csv')
        vis_loss(loss_df,output_dir)
        pbar.close()

def vis_loss(loss_df, output_dir):
    
    plot = px.line(
        data_frame=loss_df,
        x = 'step',
        y = 'loss',
        color = 'Train/Val'
    )

    plot.write_image(output_dir+'Loss_Plot.png')

def test(model_path, output_size, holdout_data, n):

    model = CellClassificationModel(
        output_size = output_size,
        simple=False
    )

    if torch.cuda.is_available():
        device = torch.device('cuda')
        model.load_state_dict(torch.load(model_path,weights_only=True))
    else:
        device = torch.device('cpu')
        model.load_state_dict(torch.load(model_path,weights_only=True,map_location = torch.device('cpu')))

    model.to(device)
    model.eval()

    test_dataloader = iter(DataLoader(holdout_data,batch_size=1,shuffle=False))

    with torch.no_grad():
        for idx in range(n):
            image,gt = next(test_dataloader)
            pred = model(image.to(device)).detach().cpu().numpy()

            image = np.moveaxis(np.squeeze(image.detach().cpu().numpy()),source=0,destination=-1)

            plot = go.Figure(
                px.imshow(image)
            )

            print(f'Predicted: {pred}, GT: {gt.cpu().numpy()}')



# %%
slides = [
    "40775.tif"
]
annotations = load_histomics('Cells.json')[0]

# Splitting the data into training and validation sets:
total_anns = len(annotations['features'])
train_test_split = 0.75
train_annotations = [{
    'type': 'FeatureCollection',
    'features': annotations['features'][:floor(train_test_split*total_anns)][0:15000],
    'properties': {'name': 'Train Cells'}
}]

val_annotations = [{
    'type': 'FeatureCollection',
    'features': annotations['features'][floor(train_test_split*total_anns):][0:8000],
    'properties': {'name': 'Validation Cells'}
}]

batch_size = 16

# Image augmentations (Normalization means and std are optimized for ImageNet pre-trained models)
train_transforms = Compose([
    RandomHorizontalFlip(p=0.5),
    RandomVerticalFlip(p=0.5),
    ColorJitter(),
    ToTensor(),
    Normalize(mean = [0.485, 0.456, 0.406], std = [0.229,0.224,0.225])
])

val_transforms = Compose([
    ToTensor(),
    Normalize(mean = [0.485,0.456, 0.406], std = [0.229, 0.224, 0.225])
])

# Converts list of labels per cell to proportions of each cell type
label_transform = lambda imm_list: cell_percentages(imm_list)

print('Starting dataset construction')
train_data = ClassificationDataset(
    slides = slides,
    annotations = train_annotations,
    label_property = 'Main_Cell_Types --> IMM',
    transforms = train_transforms,
    label_transforms = label_transform,
    use_cache = False,
    use_parallel=False,
    patch_mode = 'centered_bbox',
    verbose = True
)

val_data = ClassificationDataset(
    slides = slides,
    annotations = val_annotations,
    label_property = 'Main_Cell_Types --> IMM',
    transforms = val_transforms,
    label_transforms = label_transform,
    use_cache = False,
    use_parallel=False,
    patch_mode = 'centered_bbox',
    verbose = True
)

print(f'size of train_data: {sys.getsizeof(train_data.data)}, size of val_data: {sys.getsizeof(val_data.data)}')
print('Datasets prepared!')

train_dataloader = DataLoader(train_data, batch_size = batch_size, shuffle = True)
val_dataloader = DataLoader(val_data, batch_size = batch_size, shuffle = True)


# %%

model = CellClassificationModel(
    output_size = 2,
    simple = False
)

optimizer = torch.optim.Adam([
    dict(params = model.parameters(), lr = 5e-5, weight_decay = 0.00001)
])

loss = torch.nn.CrossEntropyLoss()


# %%
print('Starting Training!')
print(f'Number of steps per epoch: {round(len(train_data)/batch_size)}')
train(
    train_data = train_dataloader,
    val_data = val_dataloader,
    model = model,
    optimizer = optimizer,
    loss = loss,
    output_dir = './outputs/'
)


# %%
print('Starting Testing!')
test(
    model_path = './outputs/Classification_Model.pth',
    output_size = 2,
    holdout_data = val_data,
    n = 10
)






"""Components related to DSA Plugins
"""

import requests
import json
import numpy as np
import uuid

from typing_extensions import Union
import girder_client


# Dash imports
import dash
dash._dash_renderer._set_react_version('18.2.0')
from dash import dcc, callback, ctx, ALL, MATCH, exceptions, no_update
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
from dash_extensions.enrich import DashBlueprint, html, Input, Output, State, PrefixIdTransform, MultiplexerTransform

from fusion_tools.visualization.vis_utils import get_pattern_matching_value
from fusion_tools import DSATool
from fusion_tools.handler.resource_selector import DSAResourceSelector

from girder_job_sequence import Job, Sequence
from girder_job_sequence.utils import from_list, from_dict
import threading

PARAMETER_TAGS = ['integer','float','double','boolean','string','integer-vector','float-vector','double-vector','string-vector',
                'integer-enumeration','float-enumeration','double-enumeration','string-enumeration','file','directory','image',
                'geometry','point','pointfile','region','table','transform']

class DSAPluginGroup(DSATool):
    """Blueprint object which creates input layout for sets of plugins/sequences of plugins

    :param DSATool: Class for components that integrate with DSA
    :type DSATool: None
    """
    def __init__(self,
                 handler:None
                ):
        
        super().__init__()
        self.handler = handler

        self.embedded_ids = [
            'dsa-resource-selector'
        ]

        self.base_id = 'dsa-plugin'

    def __str__(self):
        return 'DSA Plugin Group'

    def load(self, component_prefix: int):
        
        self.component_prefix = component_prefix

        self.title = 'DSA Plugin Group'
        self.blueprint = DashBlueprint(
            transforms=[
                PrefixIdTransform(prefix=f'{component_prefix}',escape = lambda input_id: self.prefix_escape(input_id)),
                MultiplexerTransform()
            ]
        )

        self.get_callbacks()

        # Adding DSAResourceSelector to DSAPluginGroup
        self.resource_selector = DSAResourceSelector(
            handler = self.handler
        )
        self.resource_selector.load(self.component_prefix)

        #TODO: Either add a region selector to DSAResourceSelector or create a separate region selector component
    
    def load_plugin(self, plugin_dict:dict, session_data:dict, component_index:int, sequence=False):
        
        exe_dict = None
        self.handler.gc.setToken(
            session_data['current_user']['token']
        )

        try:
            job_obj = Job(
                self.handler.gc,
                plugin_id = plugin_dict['_id'] if '_id' in plugin_dict else None,
                docker_image = plugin_dict['image'] if 'image' in plugin_dict else None,
                cli = plugin_dict['name'] if 'name' in plugin_dict else None,
                input_args = plugin_dict['input_args'] if 'input_args' in plugin_dict else []
            )

            exe_dict = job_obj.executable_dict
        except girder_client.HttpError:
            # Error getting this plugin
            return dbc.Alert(f'Error getting plugin: {plugin_dict}',color='danger')
        
        if 'input_args' in plugin_dict:
            # Parsing through the provided input_args and pulling them out of the plugin_parameters
            inputs_list = []
            for in_arg in plugin_dict['input_args']:
                if type(in_arg)==str:
                    # Looking for the input with this name and setting default from input (if specified)
                    exe_input = self.find_executable_input(exe_dict,in_arg)
                    exe_input['disabled'] = False
                elif type(in_arg)==dict:
                    # Looking for the input with in_arg['name'] and setting default from in_arg
                    exe_input = self.find_executable_input(exe_dict, in_arg['name'])
                    if exe_input is None:
                        # For "output" channel files, {parameter_name}_folder is also needed but won't be specified in the XML
                        exe_input = {
                            'name': in_arg['name'],
                            'label': in_arg['name'],
                            'description': 'Output file folder',
                            'type': 'directory'
                        }
                    
                    exe_input['disabled'] = in_arg['disabled'] if 'disabled' in in_arg else False

                    if 'default' in in_arg:
                        if type(in_arg['default']) in [int,float,str]:
                            exe_input['default'] = in_arg['default']
                        elif type(in_arg['default'])==dict:
                            # Defining input from uploaded items/file id
                            if 'value' in in_arg['default']:
                                exe_input['default'] = in_arg['default']
                            
                            elif in_arg['default']['type'] in ['upload_folder','upload_item']:
                                # Find uploaded_items:
                                uploaded_items = [i for i in session_data['uploaded_files'] if i['_modelType']=='file']
                                # Which one matches the default name
                                uploaded_item_names = [i['fusion_upload_name'] for i in uploaded_items]
                                if in_arg['default']['name'] in uploaded_item_names:
                                    matching_item = uploaded_items[uploaded_item_names.index(in_arg['default']['name'])]
                                    matching_item_info = self.handler.get_item_info(matching_item['itemId'],session_data['current_user']['token'])
                                    if in_arg['default']['type']=='upload_folder':
                                        folder_info = self.handler.get_folder_info(matching_item_info['folderId'],session_data['current_user']['token'])
                                        exe_input['default'] = {
                                            'name': folder_info['name'],
                                            '_id': folder_info['_id']
                                        }
                                    elif in_arg['default']['type']=='upload_item':
                                        exe_input['default'] = {
                                            'name': matching_item_info['name'],
                                            '_id': matching_item_info['_id']
                                        }
                                    
                                else:
                                    exe_input['default'] = None
                            
                            elif in_arg['default']['type']=='upload_file':
                                uploaded_files = [i for i in session_data['uploaded_files'] if i['_modelType']=='file']
                                uploaded_file_names = [i['fusion_upload_name'] for i in uploaded_files]
                                if in_arg['default']['name'] in uploaded_file_names:
                                    matching_file = uploaded_files[uploaded_file_names.index(in_arg['default']['name'])]
                                    exe_input['default'] = {
                                        'name': matching_file['name'],
                                        '_id': matching_file['_id']
                                    }
                                else:
                                    exe_input['default'] = None
                            
                            elif in_arg['default']['type']=='intermediate_file':
                                # This is for creating a wildcard input
                                if 'wildcard' in in_arg['default']:
                                    exe_input['default'] = {
                                        'name': in_arg['default']['wildcard'],
                                        '_id': in_arg['default']['wildcard']
                                    }
                                if 'transform' in in_arg['default']:
                                    base = in_arg['default']['transform']['base']
                                    uploaded_files = [i for i in session_data['uploaded_files'] if i['_modelType']=='file']
                                    uploaded_file_names = [i['fusion_upload_name'] for i in uploaded_files]
                                    if base in uploaded_file_names:
                                        matching_file = uploaded_files[uploaded_file_names.index(base)]

                                        if 'ext' in in_arg['default']['transform']:
                                            old_ext = matching_file['name'].split('.')[-1]
                                            new_name = matching_file['name'].replace(f'.{old_ext}',in_arg['default']['transform']['ext'])

                                            exe_input['default'] = {
                                                'name': new_name,
                                                '_id': "{{'type':'file','item_type':'_id','item_query':'"+matching_file['itemId']+"','file_type': 'fileName','file_query':'"+new_name+"'}}"
                                            }

                                        elif 'replace' in in_arg['default']['transform']:
                                            replace_str = in_arg['default']['transform']['replace']
                                            new_name = matching_file['name'].replace(replace_str[0],replace_str[1])

                                            exe_input['default'] = {
                                                'name': new_name,
                                                '_id': "{{'type':'file','item_type':'_id','item_query':'"+matching_file['itemId']+"','file_type':'fileName','file_query':'"+new_name+"'}}"
                                            }

                                            
                                        else:
                                            exe_input['default'] = None

                                    else:
                                        exe_input['default'] = None

                            elif in_arg['default']['type']=='upload_annotation':
                                uploaded_annotations = [i for i in session_data['uploaded_files'] if i['_modelType']=='annotation']

                                if 'fileName' in in_arg['default']:
                                    uploaded_parent_names = [i['parentName'] for i in uploaded_annotations]
                                    if in_arg['default']['fileName'] in uploaded_parent_names:
                                        matching_parent = uploaded_annotations[uploaded_parent_names.index(in_arg['default']['name'])]
                                        annotation_info = self.handler.get_annotation_names(matching_parent['_id'], user_token = session_data['current_user']['token'])

                                        if 'annotationName' in in_arg['default']:
                                            annotation_names = [i['annotation']['name'] for i in annotation_info]
                                            if in_arg['default']['annotationName'] in annotation_names:
                                                matching_annotation = annotation_info[annotation_names.index(in_arg['default']['annotationName'])]
                                                exe_input['default'] = {
                                                    'name': matching_annotation['annotation']['name'],
                                                    '_id': matching_annotation['_id']
                                                }
                                            else:
                                                exe_input['default'] = None
                                        elif 'annotationId' in in_arg['default']:
                                            annotation_ids = [i['_id'] for i in annotation_info]
                                            if in_arg['default']['annotationId'] in annotation_ids:
                                                matching_annotation = annotation_info[annotation_ids.index(in_arg['default']['annotationId'])]
                                                exe_input['default'] = {
                                                    'name': matching_annotation['annotation']['name'],
                                                    '_id': matching_annotation['_id']
                                                }
                                            else:
                                                exe_input['default'] = None
                                    else:
                                        exe_input['default'] = None

                                elif 'fileId' in in_arg['default']:
                                    uploaded_parent_ids = [i['_id'] for i in uploaded_annotations]
                                    if in_arg['default']['fileId'] in uploaded_parent_ids:
                                        matching_parent = uploaded_annotations[uploaded_parent_ids.index(in_arg['default']['fileId'])]
                                        annotation_info = self.handler.get_annotation_names(matching_parent['_id'], user_token = session_data['current_user']['token'])

                                        if 'annotationName' in in_arg['default']:
                                            annotation_names = [i['annotation']['name'] for i in annotation_info]
                                            if in_arg['default']['annotationName'] in annotation_names:
                                                matching_annotation = annotation_info[annotation_names.index(in_arg['default']['annotationName'])]
                                                exe_input['default'] = {
                                                    'name': matching_annotation['annotation']['name'],
                                                    '_id': matching_annotation['_id']
                                                }
                                            else:
                                                exe_input['default'] = None
                                        elif 'annotationId' in in_arg['default']:
                                            annotation_ids = [i['_id'] for i in annotation_info]
                                            if in_arg['default']['annotationId'] in annotation_ids:
                                                matching_annotation = annotation_info[annotation_ids.index(in_arg['default']['annotationId'])]
                                                exe_input['default'] = {
                                                    'name': matching_annotation['annotation']['name'],
                                                    '_id': matching_annotation['_id']
                                                }
                                            else:
                                                exe_input['default'] = None
                                    else:
                                        exe_input['default'] = None

                                else:
                                    exe_input['default'] = None

                            elif in_arg['default']['type']=='output_file':
                                # This will need fileName and folderId keys
                                output_file_name = in_arg['default']['fileName']
                                output_folder_id = in_arg['default']['folderId']

                                exe_input['default'] = {}
                                if type(output_file_name)==str:
                                    exe_input['default']['fileName'] = output_file_name
                                elif type(output_file_name)==dict:
                                    reference = output_file_name['name']
                                    uploaded_items = [i for i in session_data['uploaded_files'] if i['_modelType']=='file']
                                    # Which one matches the default name
                                    uploaded_item_names = [i['fusion_upload_name'] for i in uploaded_items]
                                    if reference in uploaded_item_names:
                                        matching_item = uploaded_items[uploaded_item_names.index(reference)]
                                        matching_item_info = self.handler.get_item_info(matching_item['itemId'],session_data['current_user']['token'])

                                        item_name = matching_item_info['name']
                                        # Could do something with this, add a _output.{ref_ext} or something if a new extension isn't provided
                                        ref_ext = item_name.split('.')[-1]
                                        if 'ext' in output_file_name:
                                            new_ext = output_file_name['ext'] if '.' in output_file_name['ext'] else f'.{output_file_name["ext"]}'
                                            if '.' in item_name:
                                                exe_input['default']['fileName'] = '.'.join(item_name.split('.')[:-1])+new_ext
                                            else:
                                                exe_input['default']['fileName'] = item_name+new_ext
                                        else:
                                            exe_input['default']['fileName'] = item_name

                                if type(output_folder_id)==str:
                                    exe_input['default']['folderId'] = output_folder_id
                                elif type(output_folder_id)==dict:
                                    reference = output_folder_id['name']
                                    uploaded_items = [i for i in session_data['uploaded_files'] if i['_modelType']=='file']
                                    # Which one matches the default name
                                    uploaded_item_names = [i['fusion_upload_name'] for i in uploaded_items]
                                    if reference in uploaded_item_names:
                                        matching_item = uploaded_items[uploaded_item_names.index(reference)]
                                        matching_item_info = self.handler.get_item_info(matching_item['itemId'],session_data['current_user']['token'])
                                        folder_info = self.handler.get_folder_info(matching_item_info['folderId'],session_data['current_user']['token'])

                                        exe_input['default']['folderId'] = folder_info['_id']
                                    else:
                                        exe_input['default'] = None
                                else:
                                    exe_input['default'] = None

                            else:
                                exe_input['default'] = None
                        else:
                            exe_input['default'] = None
                
                if not sequence:
                    inputs_list.append(exe_input | {'plugin_id': job_obj.plugin_id, 'unique_plugin_id': plugin_dict["unique_plugin_id"]})
                else:
                    inputs_list.append(exe_input | {'plugin_id': job_obj.plugin_id, 'unique_plugin_id': plugin_dict["unique_plugin_id"], 'sequence_id': plugin_dict['sequence_id']})
        else:
            inputs_list = []
            for ip in exe_dict['parameters']:
                # ip = input, ipp = input parameter
                if not sequence:
                    inputs_list.extend([ipp | {'disabled': False, 'plugin_id': job_obj.plugin_id, 'unique_plugin_id': plugin_dict['unique_plugin_id']} for ipp in ip['inputs']])
                else:
                    inputs_list.extend([ipp | {'disabled': False, 'plugin_id': job_obj.plugin_id, 'unique_plugin_id': plugin_dict['unique_plugin_id'], 'sequence_id': plugin_dict['sequence_id']} for ipp in ip['inputs']])


        # Now creating the interactive component (without component-prefix, can transform later)
        # https://stackoverflow.com/questions/21716940/is-there-a-way-to-track-the-number-of-times-a-function-is-called
        file_input_idx = [0]
        plugin_component = html.Div([
            dcc.Store(
                id = {'type': 'dsa-plugin-plugin-info-store','index': component_index} if not sequence else {'type': 'dsa-plugin-sequence-plugin-info-store','index': component_index},
                data = json.dumps({
                    '_id': job_obj.plugin_id,
                    'exe_dict': exe_dict.copy(),
                    'input_args': job_obj.input_args,
                    'unique_plugin_id': plugin_dict['unique_plugin_id'],
                    'sequence_id': plugin_dict['sequence_id'] if 'sequence_id' in plugin_dict else None
                }),
                storage_type = 'memory'
            ),
            dbc.Row([
                html.H5(
                    html.A(
                        exe_dict['title'],
                        target = '_blank',
                        href = exe_dict['documentation'] if 'documentation' in exe_dict else None
                    )
                )
            ]),
            html.Hr(),
            dbc.Row([
                exe_dict['description']
            ]),
            dbc.Row([
                dmc.AvatarGroup(
                    children = [
                        dmc.Tooltip(
                            dmc.Avatar(
                                ''.join([n[0] for n in author.split() if not n[0] in ['(',')']]),
                                size = 'lg',
                                radius = 'xl',
                                color = f'rgb({np.random.randint(0,255)},{np.random.randint(0,255)},{np.random.randint(0,255)})'
                            ),
                            label = author,
                            position = 'bottom'
                        )
                        for author in exe_dict['author'].split(',')
                    ]
                )
            ]),
            html.Hr(),
            html.Div(
                dbc.Stack([
                    self.make_input_component(inp,inp_idx, file_input_idx,sequence = sequence)
                    for inp_idx, inp in enumerate(inputs_list)
                ],direction='vertical',gap=2),
                style = {'maxHeight': '80vh','overflow':'scroll'}
            ),
            dbc.Button(
                'Submit Plugin',
                className = 'd-grid col-12 mx-auto',
                color = 'success',
                disabled = False,
                id = {'type': 'dsa-plugin-submit-button','index': component_index}
            ) if not sequence else html.Div()
        ])

        return plugin_component

    def find_executable_input(self, executable_dict, input_name)->dict:

        exe_input = None
        for p in executable_dict['parameters']:
            for inp in p['inputs']:
                if inp['name']==input_name:
                    exe_input = inp
                    break
        
        return exe_input

    def make_input_component(self, input_dict, input_index, file_input_index, sequence):
        
        input_desc_column = [
            dbc.Row(html.H6(input_dict['label'])),
            dbc.Row(html.P(input_dict['description'])),
            dcc.Store(
                id = {'type': 'dsa-plugin-plugin-input-info','index': input_index} if not sequence else {'type': 'dsa-plugin-sequence-input-info','index': input_index},
                data = json.dumps(input_dict),
                storage_type = 'memory'
            )
        ]

        if 'enumeration' in input_dict['type']:
            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        dcc.Dropdown(
                            options = [
                                {'label': i, 'value': i}
                                for i in input_dict['options']
                            ],
                            multi = False,
                            disabled = input_dict['disabled'],
                            value = input_dict['default'] if not input_dict['default'] is None else [],
                            id = {'type': 'dsa-plugin-input','index': input_index} if not sequence else {'type': 'dsa-plugin-sequence-input','index': input_index},
                            style = {'width': '100%'}
                        )
                    ],md=7)
                ]),
                html.Hr()
            ])
        elif input_dict['type'] in ['region','geometry','point']:

            #TODO: Replace this with a real region/geometry selector
            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        dbc.Input(
                            type = 'text',
                            value = input_dict['default'] if not input_dict['default'] is None else [],
                            maxLength = 1000,
                            disabled = input_dict['disabled'],
                            id = {'type': 'dsa-plugin-input','index': input_index} if not sequence else {'type': 'dsa-plugin-sequence-input','index': input_index},
                            style = {'width': '100%'}
                        )
                    ],md=7)
                ]),
                html.Hr()
            ])

        elif input_dict['type'] in ['file','directory','image']:
            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        self.make_file_component(
                            select_type = input_dict['type'],
                            channel = input_dict['channel'] if 'channel' in input_dict else 'input',
                            value = input_dict['default'] if 'default' in input_dict else "",
                            component_index = input_index,
                            file_selector_index = file_input_index[0],
                            disabled = input_dict['disabled'],
                            sequence = sequence
                        )
                    ], md = 7)
                ]),
                html.Hr()
            ])      

            file_input_index[0]+=1

        elif input_dict['type']=='boolean':
            # This input type cannot be disabled
            if input_dict['default'] is None:
                input_dict['default'] = False

            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        dcc.RadioItems(
                            options = [
                                {'label': 'True', 'value': 1},
                                {'label': 'False','value': 0}
                            ],
                            value = 1 if input_dict['default'] else 0,
                            id = {'type': 'dsa-plugin-input','index': input_index} if not sequence else {'type': 'dsa-plugin-sequence-input','index': input_index}
                        )
                    ],md=7)
                ]),
                html.Hr()
            ])
        elif input_dict['type'] in ['integer','float','string','double'] or 'vector' in input_dict['type']:
            if not 'constraints' in input_dict:
                input_dict['constraints'] = None

            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        dbc.Input(
                            type = 'text' if input_dict['type']=='string' else 'number',
                            value = input_dict['default'] if not input_dict['default'] is None else [],
                            maxLength = 1000,
                            disabled = input_dict['disabled'],
                            #min = input_dict['constraints']['min'] if not input_dict['constraints'] is None else [],
                            #max = input_dict['constraints']['max'] if not input_dict['constraints'] is None else [],
                            #step = input_dict['constraints']['step'] if not input_dict['constraints'] is None else [],
                            id = {'type': 'dsa-plugin-input','index': input_index} if not sequence else {'type': 'dsa-plugin-sequence-input','index': input_index},
                            style = {'width': '100%'}
                        )
                    ],md=7)
                ]),
                html.Hr()
            ])

        return input_component

    def make_file_component(self, select_type: str, channel:str, value: Union[str,dict], component_index:int, file_selector_index: int, disabled:bool,sequence:bool=False):
        
        if type(value)==dict:
            if all([i in value for i in ['name','_id']]):
                selector_value = value['name']
                input_value = value['_id']
            elif 'value' in value:
                selector_value = value['value']
                input_value = value['value']
            elif all([i in value for i in ['fileName','folderId']]):
                selector_value = value['fileName']
                input_value = json.dumps(value)
        else:
            selector_value = value
            input_value = value
    
        file_component = html.Div([
            dcc.Store(
                id = {'type': 'dsa-plugin-resource-selector-info','index': file_selector_index},
                data = json.dumps({
                    'type': select_type,
                    'channel': channel,
                    'component_index': component_index,
                    'file_component_index': file_selector_index
                }),
                storage_type='memory'
            ),
            dbc.InputGroup([
                dbc.InputGroupText(
                    f'{select_type}: '
                ),
                dbc.Input(
                    id = {'type': 'dsa-plugin-resource-selector-input','index': file_selector_index},
                    placeholder = select_type,
                    type = 'text',
                    required = True,
                    value = selector_value,
                    maxLength = 1000,
                    disabled = True
                ),
                dbc.Input(
                    id = {'type': 'dsa-plugin-input','index': component_index} if not sequence else {'type': 'dsa-plugin-sequence-input','index': component_index},
                    value = input_value,
                    style = {'display': 'none'}
                ),
                dbc.Button(
                    children = [
                        html.I(
                            className = 'fa-solid fa-file'
                        )
                    ],
                    id = {'type': 'dsa-plugin-resource-selector-open-modal','index': file_selector_index} if not sequence else {'type': 'dsa-plugin-sequence-resource-selector-open-modal','index': component_index},
                    disabled = disabled
                )
            ])
        ])

        return file_component

    def update_layout(self, session_data: dict, use_prefix:bool, plugin_groups: Union[list,dict,None]=None):
        """Updating layout of the plugin group blueprint, creates accordions for each singular plugin and outlined accordions corresponding to sequences.
        For more information on plugin sequences, see: https://github.com/spborder/girder-job-sequence

        :param session_data: Current session data, includes current user information ('_id', 'token')
        :type session_data: dict
        :param use_prefix: Used to indicate whether or not this is an initialization (False) of the layout or just updating components (True)
        :type use_prefix: bool
        :param plugin_groups: Either a list of plugins and sequence(s) or dictionary for single plugin, defaults to None
        :type plugin_groups: Union[list,dict,None], optional
        """

        plugin_components = []
        if not plugin_groups is None:
            if type(plugin_groups)==dict:
                plugin_groups = [plugin_groups]
            
            plugin_count = 0
            sequence_plugin_count = 0
            sequence_count = 0
            for p_idx, p in enumerate(plugin_groups):
                if type(p)==dict:
                    # This is a singular job that can be executed independently of any other processing plugin
                    unique_plugin_id = uuid.uuid4().hex[:24]
                    p_component = self.load_plugin(
                        plugin_dict = p | {'unique_plugin_id': unique_plugin_id},
                        session_data = session_data,
                        component_index = plugin_count,
                        sequence=False
                    )

                    plugin_components.append(
                        dmc.AccordionItem([
                            dmc.AccordionControl(p['name']),
                            dmc.AccordionPanel(
                                dbc.Stack([
                                    p_component,
                                    html.Div(
                                        id = {'type': 'dsa-plugin-submit-status-div','index': plugin_count},
                                        children = []
                                    )
                                ])
                            )
                        ],value=f'dsa-plugin-{plugin_count}',style={'marginBottom':'10px','marginTop': '10px'})
                    )
                    plugin_count += 1
                elif type(p)==list:
                    # This is a sequential job where multiple jobs are submitted to a sequence handler at the same time and executed one-by-one
                    sequence_color = f'rgb({np.random.randint(0,255)},{np.random.randint(0,255)},{np.random.randint(0,255)})'
                    sequence_id = uuid.uuid4().hex[:24]
                    
                    sequence_components = []
                    plugin_identifiers = []
                    for j_idx,j in enumerate(p):
                        unique_plugin_id = uuid.uuid4().hex[:24]

                        j_component = self.load_plugin(
                            plugin_dict = j | {'sequence_id': sequence_id, 'unique_plugin_id': unique_plugin_id},
                            session_data = session_data,
                            component_index = sequence_plugin_count,
                            sequence=True
                        )

                        sequence_components.append(
                            dmc.AccordionItem([
                                dmc.AccordionControl(j['name']),
                                dmc.AccordionPanel(
                                    dbc.Stack([
                                        j_component
                                    ])
                                )
                            ],value = f'dsa-plugin-sequence-{sequence_count}-{j_idx}')
                        )
                        sequence_plugin_count += 1
                        plugin_identifiers.append(unique_plugin_id)
                    
                    sequence_components.append(
                        dmc.AccordionItem([
                            dmc.AccordionControl(f'Submit Sequence {sequence_count+1}'),
                            dmc.AccordionPanel(
                                dbc.Stack([
                                    dbc.Button(
                                        'Submit Sequence',
                                        id = {'type': 'dsa-plugin-sequence-submit-button','index': sequence_count},
                                        className = 'd-grid col-12 mx-auto',
                                        n_clicks = 0,
                                        color = 'success',
                                        disabled = False
                                    ),
                                    dcc.Store(
                                        id = {'type': 'dsa-plugin-sequence-sequence-info','index': sequence_count},
                                        data = json.dumps({'_id': sequence_id, 'plugins': plugin_identifiers}),
                                        storage_type='memory'
                                    ),
                                    html.Div(
                                        id = {'type': 'dsa-plugin-sequence-submit-status-div','index': sequence_count},
                                        children = []
                                    )
                                ])
                            )
                        ], value = f'dsa-sequence-{sequence_count}-submit',style={'marginBottom':'10px'})
                    )

                    plugin_components.append(
                        dbc.Stack(
                            children = [
                                html.H5(f'Sequence {sequence_count+1} Plugins'),
                                dmc.Accordion(
                                    id = {'type': 'dsa-plugin-sequence-accordion','index': sequence_count},
                                    children = sequence_components,
                                )
                            ],
                            style = {
                                'padding': '10px 10px 10px 10px',
                                'outline-style':'dashed',
                                'outline-color': sequence_color
                            }
                        )
                    )
                    sequence_count += 1

        self.resource_selector.update_layout(session_data,use_prefix=False)

        layout = html.Div([
            dbc.Card([
                dbc.Modal(
                    id = {'type': 'dsa-plugin-resource-selector-modal','index': 0},
                    centered = True,
                    is_open = False,
                    size = 'xl',
                    className = None,
                    children = [
                        self.resource_selector.blueprint.embed(self.blueprint)
                    ]
                ),
                dbc.CardHeader(self.title),
                dbc.CardBody([
                    dbc.Row(
                        'Below are separate input components for running a desired set of plugins. Individual plugins can be executed independently while plugins that are part of a Sequence are submitted all at once.'
                    ),
                    html.Hr(),
                    html.Div(
                        id = {'type': 'dsa-plugin-plugin-parent-div','index': 0},
                        children = dmc.Accordion(
                            children = plugin_components
                        )
                    )
                ])
            ])
        ],style = {'maxHeight': '90vh','overflow': 'scroll'})

        if use_prefix:
            PrefixIdTransform(prefix=f'{self.component_prefix}', escape = lambda input_id: self.prefix_escape(input_id)).transform_layout(layout)

        return layout

    def gen_layout(self, session_data:dict):
        """Initializing layout of blueprint.

        :param session_data: Dictionary containing information from current session
        :type session_data: dict
        """

        self.blueprint.layout = self.update_layout(session_data, use_prefix=False)

    def get_callbacks(self):
        """Registering callbacks with blueprint object.
        """

        # Callback for running individual plugin
        self.blueprint.callback(
            [
                Input({'type': 'dsa-plugin-submit-button','index': MATCH},'n_clicks'),
            ],
            [
                State({'type': 'dsa-plugin-plugin-info-store','index': ALL},'data'),
                State({'type': 'dsa-plugin-input','index': ALL},'value'),
                State({'type': 'dsa-plugin-plugin-input-info','index': ALL},'data'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-plugin-submit-status-div','index': MATCH},'children'),
                Output({'type': 'dsa-plugin-submit-button','index': MATCH},'disabled'),
            ],
            prevent_initial_call = True
        )(self.submit_plugin)

        # Callback for running sequence
        self.blueprint.callback(
            [
                Input({'type': 'dsa-plugin-sequence-submit-button','index': MATCH},'n_clicks')
            ],
            [
                State({'type': 'dsa-plugin-sequence-sequence-info','index': MATCH},'data'),
                State({'type': 'dsa-plugin-sequence-plugin-info-store','index': ALL},'data'),
                State({'type': 'dsa-plugin-sequence-input','index': ALL},'value'),
                State({'type': 'dsa-plugin-sequence-input-info','index': ALL},'data'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-plugin-sequence-submit-status-div','index': MATCH},'children'),
                Output({'type': 'dsa-plugin-sequence-submit-button','index': MATCH},'disabled')
            ],
            prevent_initial_call = True
        )(self.submit_sequence)

        # Callback for opening the resource selector modal with information
        self.blueprint.callback(
            [
                Input({'type': 'dsa-plugin-resource-selector-open-modal','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'dsa-plugin-resource-selector-info','index': ALL},'data'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-plugin-resource-selector-modal','index': ALL},'is_open'),
                Output({'type': 'dsa-plugin-resource-selector-modal','index': ALL},'children')
            ],
            prevent_initial_call = True
        )(self.open_resource_selector)

        # Callback for retrieving selected resource information
        self.blueprint.callback(
            [
                Input({'type': 'dsa-plugin-resource-selector-modal','index': ALL},'is_open')
            ],
            [
                State({'type': 'anchor-dsa-resource-selector-selected-resources','index': ALL},'data'),
            ],
            [
                Output({'type': 'dsa-plugin-resource-selector-input','index': ALL},'value'),
                Output({'type': 'dsa-plugin-input','index': ALL},'value')
            ],
            prevent_initial_call = True
        )(self.retrieve_resource_selection)

        # Callback for creating a wildcard input
        """
        self.blueprint.callback(
            [
                Input({'type': 'dsa-plugin-wildcard-icon','index': ALL},'n_clicks')
            ],
            [
                State({'type': '','index': },'')
            ],
            [
                Output({'type': '', 'index': },'')
            ],
            prevent_initial_call = True
        )(self.create_wildcard_input)

        """

    def post_process_plugin_inputs(self, plugin_inputs, plugin_input_info):

        processed_inputs = []
        error_inputs = []
        for p_input, p_info in zip(plugin_inputs,plugin_input_info):
            # None type inputs are replaced with default if one is available
            if not p_input is None:
                if p_info['type'] in ['file','directory','image'] and p_info['channel']=='output':
                    if '"' in p_input or "'" in p_input:
                        p_input = json.loads(p_input.replace("'",'"'))
                    if type(p_input)==list:
                        p_input = p_input[0]
                    if 'channel' in p_info:
                        if p_info['channel']=='output':
                            # This needs a _name and _folder argument
                            if type(p_input)==dict:
                                processed_inputs.extend([
                                    {
                                        'name': p_info['name'],
                                        'value':p_input['fileName']
                                    },
                                    {
                                        'name': p_info['name']+'_folder',
                                        'value':p_input['folderId']
                                    }
                                ])

                            else:
                                error_inputs.append(p_info['name'])
                        else:
                            if type(p_input)==str:
                                processed_inputs.append({
                                    'name': p_info['name'],
                                    'value':p_input
                                })
                            else:
                                error_inputs.append(p_info['name'])

                    else:
                        if type(p_input)==str:
                            processed_inputs.append({
                                'name': p_info['name'],
                                'value':p_input
                            })
                        else:
                            error_inputs.append(p_info['name'])

                elif p_info['type'] in ['region','geometry','point']:
                    
                    if p_info['type']=='region':
                        if not all([j in p_input for j in ['[',']']]):
                            fixed_region = p_input.replace(' ','').split(',')
                        else:
                            fixed_region = p_input.replace('[','').replace(']','').replace(' ','').split(',')

                        try:
                            fixed_region = [float(i) for i in fixed_region]
                            processed_inputs.append({
                                'name': p_info['name'],
                                'value':json.dumps(fixed_region)
                            })
                        except ValueError:
                            error_inputs.append(p_info['name'])

                    else:
                        error_inputs.append(p_info['name'])

                elif 'vector' in p_info['type']:
                    
                    fixed_vector = p_input.replace('[','').replace(']','').replace(' ','').split(',')
                    if not 'string' in p_info['type']:
                        try:
                            if any([i in p_info['type'] for i in ['float','double']]):
                                fixed_vector = [float(i) for i in fixed_vector]
                            elif 'integer' in p_info['type']:
                                fixed_vector = [int(i) for i in fixed_vector]
                        except ValueError:
                            error_inputs.append(p_info['name'])

                    processed_inputs.append({
                        'name': p_info['name'],
                        'value':json.dumps(fixed_vector)
                    })

                else:
                    processed_inputs.append({
                        'name': p_info['name'],
                        'value': p_input
                    })

        return processed_inputs, error_inputs

    def submit_plugin(self, clicked, plugin_info, plugin_inputs,plugin_input_info, session_data):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        session_data = json.loads(session_data)
        plugin_info = [json.loads(p_info) for p_info in plugin_info]
        all_plugin_input_info = [json.loads(i_info) for i_info in plugin_input_info]

        selected_plugin = plugin_info[ctx.triggered_id['index']]
        
        selected_plugin_inputs = []
        selected_plugin_inputs_info = []
        for p,p_info in zip(plugin_inputs,all_plugin_input_info):
            if p_info['unique_plugin_id']==selected_plugin['unique_plugin_id']:
                selected_plugin_inputs.append(p)
                selected_plugin_inputs_info.append(p_info)

        processed_inputs_dict, error_inputs = self.post_process_plugin_inputs(selected_plugin_inputs,selected_plugin_inputs_info)

        if len(selected_plugin['input_args'])>len(plugin_inputs):
            # Mismatch in input length
            status_div = dbc.Alert('Missing plugin inputs!',color='warning')
            button_disable = False
            return status_div, button_disable

        if len(error_inputs)>0:
            status_div = dbc.Alert(f'Incorrect input types for: {error_inputs}',color='warning')
            button_disable = False
            return status_div, button_disable

        job_dict = {
            'plugin_id': selected_plugin['_id'],
            'input_args': processed_inputs_dict
        }

        # function incorporated from girder-job-sequence
        job_obj = from_dict(self.handler.gc,job_dict)
        job_start_response = job_obj.start()

        if job_start_response.status_code==200:
            status_div = dbc.Alert('Plugin successfully submitted!',color='success')
            button_disable = True
        else:
            status_div = dbc.Alert(f'Error submitting plugin: {selected_plugin["_id"]}',color = 'danger')
            print(job_start_response.content)
            button_disable = False

        return status_div, button_disable

    def submit_sequence(self, clicked, sequence_info, sequence_plugins_info, sequence_plugin_inputs, sequence_plugin_inputs_info, session_data):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate

        session_data = json.loads(session_data)
        # Getting information on this sequence
        sequence_info = json.loads(sequence_info)
        # Getting information on plugins in this sequence
        all_sequence_plugins_info = [json.loads(i) for i in sequence_plugins_info]
        sequence_plugin_inputs_info = [json.loads(i) for i in sequence_plugin_inputs_info]

        selected_sequence_plugins_info = [i for i in all_sequence_plugins_info if i['sequence_id']==sequence_info['_id'] and i['unique_plugin_id'] in sequence_info['plugins']]
        selected_sequence_plugins_ids = [i['unique_plugin_id'] for i in selected_sequence_plugins_info]
        sequence_job_list = []
        sequence_inputs_error_list = []
        for plugin_id in sequence_info['plugins']:
            plugin_info = selected_sequence_plugins_info[selected_sequence_plugins_ids.index(plugin_id)]

            plugin_inputs = []
            plugin_inputs_info = []
            for s,s_info in zip(sequence_plugin_inputs, sequence_plugin_inputs_info):
                if s_info['unique_plugin_id']==plugin_id:
                    plugin_inputs.append(s)
                    plugin_inputs_info.append(s_info)

            processed_plugin_inputs, error_inputs = self.post_process_plugin_inputs(plugin_inputs,plugin_inputs_info)
            sequence_inputs_error_list.extend(error_inputs)
            
            plugin_dict = {
                'plugin_id': plugin_info['_id'],
                'input_args': processed_plugin_inputs
            }
            sequence_job_list.append(plugin_dict)
    
        if len(sequence_inputs_error_list) > 0:
            status_div = dbc.Alert(f'Incorrect input types for: {sequence_inputs_error_list}',color='warning')
            button_disable = False
            return status_div, button_disable

        sequence_obj = from_list(self.handler.gc,sequence_job_list)

        # Start job sequence thread
        job_seq_thread = threading.Thread(
            target = sequence_obj.start, 
            name = f'fusion_tools_job_sequence_{sequence_obj.id}',
            kwargs = {
                "cancel_on_error": True
            },
            daemon=True
        )
        job_seq_thread.start()

        status_div = dbc.Alert(f'Sequence {ctx.triggered_id["index"]+1} Submitted',color='success')
        button_disable = True
        return status_div, button_disable

    def open_resource_selector(self, clicked, selector_info, session_data):
        
        if any(clicked):
            opened = True
            selector_info = json.loads(selector_info[ctx.triggered_id['index']])
            session_data = json.loads(session_data)

            selector_children = self.resource_selector.update_layout(
                session_data = session_data,
                use_prefix = True,
                selector_type = selector_info['type'] if not selector_info['type']=='image' else 'item',
                select_count = 1,
                source_channel = selector_info['channel'],
                source_index = {
                    'component_index': selector_info['component_index'],
                    'file_component_index': selector_info['file_component_index']
                }
            )

            return [opened], [selector_children]

        else:
            raise exceptions.PreventUpdate
    
    def retrieve_resource_selection(self, modal_opened, selected_data):

        modal_opened = get_pattern_matching_value(modal_opened)
        selected_data = get_pattern_matching_value(selected_data)
        if selected_data and not modal_opened:
            selected_data = json.loads(selected_data)
            update_index = selected_data['source_component_index']
            selector_val = ','.join([i['Name'] for i in selected_data['resource_list']])
            input_val = json.dumps([i['_id'] if '_id' in i else i for i in selected_data['resource_list']])

            new_selector_vals = [no_update if not idx==update_index['file_component_index'] else selector_val for idx in range(len(ctx.outputs_list[0]))]
            new_input_vals = [no_update if not idx==update_index['component_index'] else input_val for idx in range(len(ctx.outputs_list[1]))]

            return new_selector_vals, new_input_vals
        else:
            raise exceptions.PreventUpdate

    def create_wildcard_input(self):
        pass





class DSAPluginRunner(DSATool):
    """Handler for DSAPluginRunner component, letting users specify input arguments to plugins to run on connected DSA instance.

    :param DSATool: Class for components that integrate with DSA.
    :type DSATool: None
    """
    def __init__(self,
                 handler: None
                 ):
        
        super().__init__()
        self.handler = handler

        self.embedded_ids = [
            'dsa-resource-selector',
            'dsa-plugin'
        ]

        self.base_id = 'dsa-plugin-runner'

    def load(self, component_prefix: int):
        
        self.component_prefix = component_prefix

        self.title = 'DSA Plugin Runner'
        self.blueprint = DashBlueprint(
            transforms=[
                PrefixIdTransform(prefix=f'{component_prefix}', escape = lambda input_id: self.prefix_escape(input_id)),
                MultiplexerTransform()
            ]
        )

        self.get_callbacks()

        # Adding DSAPluginGroup object to PluginRunner
        self.plugin_inputs_handler = DSAPluginGroup(
            handler = self.handler
        )
        self.plugin_inputs_handler.load(self.component_prefix)
    
    def update_layout(self, session_data:dict, use_prefix: bool):
        
        plugin_list = self.handler.list_plugins(user_token = session_data['current_user']['token'])
        docker_list = sorted(list(set([i['image'] for i in plugin_list])))

        if not use_prefix:
            self.plugin_inputs_handler.blueprint.layout = self.plugin_inputs_handler.update_layout(session_data,False)

        layout = html.Div([
            dbc.Card([
                dbc.CardBody([
                    dbc.Row(
                        html.H3('DSA Plugin Runner')
                    ),
                    html.Hr(),
                    dbc.Row(
                        'Select a plugin to run on the cloud!'
                    ),
                    html.Hr(),
                    dbc.Row([
                        dbc.Col([
                            dcc.Dropdown(
                                options = [
                                    {'label': i, 'value': i}
                                    for i in docker_list
                                ],
                                value = [],
                                multi = False,
                                placeholder = 'Docker Image containing Plugin',
                                id = {'type': 'dsa-plugin-runner-docker-drop','index': 0}
                            )],
                            md = 6
                        ),
                        dbc.Col([
                            dcc.Dropdown(
                                options = [],
                                value = [],
                                multi = False,
                                placeholder = 'Plugin Name',
                                id = {'type': 'dsa-plugin-runner-cli-drop','index': 0}
                            )
                        ],
                        md = 6)
                    ]),
                    html.Div(
                        id = {'type': 'dsa-plugin-runner-inputs-div','index': 0},
                        children = [
                            self.plugin_inputs_handler.update_layout(session_data,False) if use_prefix else self.plugin_inputs_handler.blueprint.embed(self.blueprint)
                        ]
                    )
                ])
            ])
        ])

        if use_prefix:
            PrefixIdTransform(prefix=f'{self.component_prefix}', escape = lambda input_id: self.prefix_escape(input_id)).transform_layout(layout)

        return layout

    def gen_layout(self, session_data:dict):
        
        self.blueprint.layout = self.update_layout(session_data, use_prefix=False)

    def get_callbacks(self):

        # Callback to get all the CLIs for a selected Docker image
        self.blueprint.callback(
            [
                Input({'type': 'dsa-plugin-runner-docker-drop','index': ALL},'value')
            ],
            [
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-plugin-runner-cli-drop','index': ALL},'options')
            ],
            prevent_initial_call = True
        )(self.update_cli_options)

        # Callback to load plugin input components from CLI selection
        self.blueprint.callback(
            [
                Input({'type': 'dsa-plugin-runner-cli-drop','index': ALL},'value')
            ],
            [
                State({'type': 'dsa-plugin-runner-docker-drop','index': ALL},'value'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-plugin-runner-inputs-div','index': ALL},'children')
            ],
            prevent_initial_call=True
        )(self.populate_plugin_inputs)

    def update_cli_options(self, docker_select,session_data):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        docker_select = get_pattern_matching_value(docker_select)
        session_data = json.loads(session_data)
        plugin_list = self.handler.list_plugins(user_token=session_data['current_user']['token'])
        included_cli = sorted([i['name'] for i in plugin_list if i['image']==docker_select])

        return [included_cli]
    
    def populate_plugin_inputs(self, cli_select, docker_select,session_data):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        session_data = json.loads(session_data)
        docker_select = get_pattern_matching_value(docker_select)
        cli_select = get_pattern_matching_value(cli_select)
        plugin_components = self.plugin_inputs_handler.update_layout(
            session_data = session_data,
            use_prefix = True,
            plugin_groups = {
                'image': docker_select,
                'name': cli_select
            }
        )

        return [plugin_components]




class DSAPluginProgress(DSATool):
    """Handler for DSAPluginProgress component, letting users check the progress of currently running or previously run plugins as well as cancellation of running plugins.

    :param Tool: General class for components that perform visualization and analysis of data.
    :type Tool: None
    """
    def __init__(self,
                 handler):
        
        super().__init__()
        self.handler = handler
        self.job_status_key = {
            '0': 'INACTIVE',
            '1': 'QUEUED',
            '2': 'RUNNING',
            '3': 'SUCCESS',
            '4': 'ERROR',
            '5': 'CANCELED'
        }
        self.job_status_color_key = {
            '0': 'rgb(176,184,178)',
            '1': 'rgb(219,224,67)',
            '2': 'rgb(67,172,224)',
            '3': 'rgb(50,168,82)',
            '4': 'rgb(224,27,27)',
            '5': 'rgb(227,27,154)'
        }
    
        self.modal_className = 'mw-100 p-5'

    def __str__(self):
        return 'DSA Plugin Progress'

    def load(self,component_prefix:int):
        
        self.component_prefix = component_prefix

        self.title = 'DSA Plugin Progress'

        self.blueprint = DashBlueprint(
            transforms=[
                PrefixIdTransform(prefix=f'{component_prefix}'),
                MultiplexerTransform()
            ]
        )

        self.get_callbacks()

    def generate_plugin_table(self, session_data:dict, offset:int, limit: int, next_clicks:int, prev_clicks:int, use_prefix:bool):

        # Getting all jobs for this user:
        user_jobs = self.handler.get_user_jobs(
            user_id = session_data['current_user']['_id'],
            user_token = session_data['current_user']['token'],
            offset = offset,
            limit = limit
        )

        job_properties = ['title','type','updated','when','status','_id']
        
        table_head = dmc.TableThead(
            dmc.TableTr(
                [
                    dmc.TableTh(i)
                    for i in job_properties
                ]
            )
        )

        if len(user_jobs)>0:
            table_rows = []
            for job_idx,job in enumerate(user_jobs):
                job_row = []
                for prop in job_properties:
                    if not prop=='status':
                        job_row.append(
                            dmc.TableTd(job[prop])
                        )
                    else:
                        job_row.append(
                            dmc.TableTd(self.job_status_key[str(job[prop])],style = {'background': self.job_status_color_key[str(job[prop])]})
                        )

                # Adding logs/cancel buttons
                job_row.append(
                    dmc.TableTd(
                        dbc.Button(
                            'Logs',
                            color = 'warning',
                            id = {'type': 'dsa-plugin-progress-get-logs','index': job_idx},
                            n_clicks = 0
                        )
                    )
                )

                job_row.append(
                    dmc.TableTd(
                        dbc.Button(
                            'Cancel',
                            color = 'danger',
                            id = {'type': 'dsa-plugin-progress-cancel-job','index': job_idx},
                            n_clicks = 0
                        )
                    )
                )

                table_rows.append(
                    dmc.TableTr(job_row, id = {'type': 'dsa-plugin-progress-table-row','index': job_idx})
                )

        else:
            if next_clicks>0:
                table_rows.append(
                    dmc.Tr([
                        'You have exceed all of the jobs for this user!'
                    ])
                )
            else:
                table_rows.append(
                    dmc.Tr([
                        'You have not run any jobs yet!'
                    ])
                )

        table_caption = dmc.TableCaption([
            dbc.Stack([
                dbc.Button(
                    'Load Previous 5 jobs',
                    color = 'secondary',
                    n_clicks = prev_clicks,
                    className = 'd-grid col-6 mx-auto',
                    id = {'type': 'dsa-plugin-progress-load-prev-jobs','index': 0}
                ),
                dbc.Button(
                    'Load Next 5 jobs',
                    color = 'primary',
                    n_clicks = next_clicks,
                    className = 'd-grid col-6 mx-auto',
                    id = {'type': 'dsa-plugin-progress-load-next-jobs','index': 0}
                )
            ],direction = 'horizontal',gap=2)
        ])

        table_body = dmc.TableTbody(table_rows)

        if use_prefix:
            PrefixIdTransform(prefix = f'{self.component_prefix}').transform_layout(table_head)
            PrefixIdTransform(prefix = f'{self.component_prefix}').transform_layout(table_body)
            PrefixIdTransform(prefix = f'{self.component_prefix}').transform_layout(table_caption)

        return [table_head, table_body, table_caption]

    def update_layout(self, session_data:dict, use_prefix:bool):
        
        if 'current_user' in session_data:
            running_plugins = self.generate_plugin_table(session_data, 0, 5, 0, 0, False)
            running_plugins = dmc.Table(
                running_plugins,
                id = {'type': 'dsa-plugin-progress-table-content','index': 0}
            )

        else:
            running_plugins = dbc.Alert(
                'You must be logged in to view running plugins!',
                color = 'warning'
            )

        layout = html.Div([
            dbc.Card(
                dbc.CardBody([
                    dbc.Row(
                        html.H3('DSA Plugin Progress')
                    ),
                    html.Hr(),
                    dbc.Row(
                        'Monitor the progress of currently running plugins.'
                    ),
                    dbc.Row(
                        html.Div(
                            running_plugins,
                            style = {'maxHeight': '40vh','overflow': 'scroll'}
                        )
                    ),
                    dbc.Row(
                        html.Div(
                            id = {'type': 'dsa-plugin-progress-logs-content','index': 0},
                            children = []
                        )
                    )
                ])
            )
        ])

        if use_prefix:
            PrefixIdTransform(prefix=self.component_prefix).transform_layout(layout)

        return layout

    def gen_layout(self,session_data:Union[dict,None]):
        
        self.blueprint.layout = self.update_layout(session_data=session_data,use_prefix=False)

    def get_callbacks(self):

        # Callback for loading next/prev jobs
        self.blueprint.callback(
            [
                Input({'type': 'dsa-plugin-progress-load-next-jobs','index': ALL},'n_clicks'),
                Input({'type': 'dsa-plugin-progress-load-prev-jobs','index': ALL},'n_clicks')
            ],
            [
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-plugin-progress-table-content','index':ALL},'children')
            ],
            prevent_initial_call = True
        )(self.load_new_jobs)

        # Callback for cancelling plugin/loading logs
        self.blueprint.callback(
            [
                Input({'type': 'dsa-plugin-progress-get-logs','index': ALL},'n_clicks'),
                Input({'type': 'dsa-plugin-progress-cancel-job','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'dsa-plugin-progress-table-row','index': ALL},'children'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-plugin-progress-logs-content','index': ALL},'children')
            ],
            prevent_initial_call = True
        )(self.get_logs_or_cancel)

    def load_new_jobs(self, next_clicked, prev_clicked, session_data):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate

        next_clicked = get_pattern_matching_value(next_clicked)
        prev_clicked = get_pattern_matching_value(prev_clicked)
        if prev_clicked>next_clicked:
            next_clicked = 0
            prev_clicked = 0
            offset = 0
        
        else:
            offset = next_clicked - prev_clicked
       
        session_data = json.loads(session_data)
        new_table_content = self.generate_plugin_table(
                                session_data = session_data,
                                offset = offset,
                                limit = 5,
                                next_clicks = next_clicked,
                                prev_clicks = prev_clicked,
                                use_prefix = True
                            )


        return [new_table_content]

    def get_logs_or_cancel(self, logs_clicked, cancel_clicked, table_rows, session_data):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        session_data = json.loads(session_data)

        if 'dsa-plugin-progress-get-logs' in ctx.triggered_id['type']:
            row_clicked = ctx.triggered_id['index']
            job_id = table_rows[row_clicked][5]['props']['children']

            job_logs = self.handler.get_specific_job(
                job_id = job_id,
                user_token = session_data['current_user']['token']
            )

            job_logs_div = html.Div(
                [
                    html.Div([html.P(i) for i in line.split('\n')])
                    for line in job_logs['log']
                ],
                style = {'maxHeight': '20vh','overflow': 'scroll'}
            )

        elif 'dsa-plugin-progress-cancel-job' in ctx.triggered_id['type']:
            row_clicked = ctx.triggered_id['index']
            job_id = table_rows[row_clicked][5]['props']['children']

            cancel_response = self.handler.cancel_job(
                job_id = job_id,
                user_token = session_data['current_user']['token']
            )

            job_logs_div = html.Div(
                'Cancel request sent! Close and re-open to see updated status'
            )


        return [job_logs_div]
    





"""Testing DatasetBuilder component
"""

import os
import sys
sys.path.append('./src/')

from fusion_tools.visualization import Visualization
from fusion_tools.components import SlideMap, MultiFrameSlideMap, ChannelMixer, OverlayOptions, BulkLabels, PropertyViewer, PropertyPlotter, FeatureAnnotation, HRAViewer
from fusion_tools.handler.dsa_handler import DSAHandler
from fusion_tools.handler.dataset_uploader import DSAUploadType
from fusion_tools.handler.survey import SurveyType

# This is an example upload type for an image and a file
simple_upload_type = DSAUploadType(
    name = 'Basic Upload',
    description = 'This is a test upload type consisting of an image and an associated file',
    input_files = [
        {
            'name': 'Image',
            'description': 'This is any image you would like to upload to the DSA instance',
            'accepted_types': ['png','jpg','svs','tiff','tif'],
            'preprocessing_plugins': None,
            'type': 'item',
            'required': True,
        },
        {
            'name': 'Associated File',
            'description': 'This is any other file type that you want added to the "files" of the uploaded image.',
            'accepted_types': ['csv','xlsx','txt'],
            'preprocessing_plugins': None,
            'type': 'file',
            'parent': 'Image',
            'required': False,
        },
        {
            'name': 'Annotation',
            'description': 'This is an annotation file that is processed and added to the uploaded image',
            'accepted_types': ['xml','json','geojson'],
            'preprocessing_plugins': None,
            'type': 'annotation',
            'parent': 'Image',
            'required': False
        }
    ],
    processing_plugins = [
        {
            'name': 'NucleiDetection',
            'image': 'dsarchive/histomicstk:latest',
            'input_args': [
                {
                    'name': 'inputImageFile',
                    'default': {
                        'type': 'upload_file',
                        'name': 'Image'
                    },
                    'disabled': True
                },
                {
                    'name': 'outputNucleiAnnotationFile_folder',
                    'default': {
                        'type': 'upload_folder',
                        'name': 'Image'
                    },
                    'disabled': True
                },
                {
                    'name': 'outputNucleiAnnotationFile',
                    'default': {
                        'type': 'output_file',
                        'fileName': {
                            'name': 'Image',
                            'ext': '.annot'
                        },
                        'folderId': {
                            'name': 'Image'
                        }
                    },
                    'disabled': True
                },
                'nuclei_annotation_format',
                'min_nucleus_area',
                'ignore_border_nuclei',
                'ImageInversionForm'
            ]
        }
    ],
    required_metadata = [
        {
            'name': 'Image Type',
            'values': ['Histology','Fluorescence','Unknown'],
            'required': True,
            'item': 'Image'
        },
        {
            'name': 'Image Label',
            'values': ['Label 1','Label 2','Label 3'],
            'required': False,
            'item': 'Image'
        },
        {
            'name':'Extra Metadata',
            'required': False,
            'item': 'Image'
        }
    ]
)

sequence_upload_type = DSAUploadType(
    name = 'Sequential Processing Upload',
    description='This plugin requires multiple plugins to be run in sequence',
    input_files = [
        {
            'name': 'Image',
            'description': 'This is any image you would like to upload to the DSA instance',
            'accepted_types': ['png','jpg','svs','tiff','tif'],
            'preprocessing_plugins': None,
            'type': 'item',
            'required': True,
        },
    ],
    processing_plugins=[
        [
            {
                'name': 'MultiCompartmentSegment',
                'image': 'samborder2256/multicomp:latest',
                'input_args': [
                    {
                        'name': 'files',
                        'default': {
                            'type':'upload_file',
                            'name': 'Image'
                        },
                        'disabled': True
                    },
                    {
                        'name': 'base_dir',
                        'default': {
                            'type': 'upload_folder',
                            'name': 'Image'
                        },
                        'disabled': True
                    },
                    {
                        'name': 'modelfile',
                        'default': {
                            'value': '648123761019450486d13dce'
                        },
                        'disabled': True
                    }
                ]
            },
            {
                'name': 'FeatureExtraction',
                'image': 'fusionplugins/general:latest',
                'input_args': [
                    {
                        'name': 'input_image',
                        'default': {
                            'type': 'upload_file',
                            'name': 'Image'
                        },
                        'disabled': True
                    },
                    {
                        'name': 'extract_sub_compartments',
                        'default': {
                            'value': True
                        },
                        'disabled': True
                    },
                    'hematoxylin_threshold',
                    'eosinophilic_threshold',
                    'hematoxylin_min_size',
                    'eosinophilic_min_size'
                ]
            }
        ],
        {
            'name': 'NucleiDetection',
            'image': 'dsarchive/histomicstk:latest',
            'input_args': [
                {
                    'name': 'inputImageFile',
                    'default': {
                        'type': 'upload_file',
                        'name': 'Image'
                    },
                    'disabled': True
                },
                {
                    'name': 'outputNucleiAnnotationFile_folder',
                    'default': {
                        'type': 'upload_folder',
                        'name': 'Image'
                    },
                    'disabled': True
                },
                {
                    'name': 'outputNucleiAnnotationFile',
                    'default': {
                        'type': 'output_file',
                        'fileName': {
                            'name': 'Image',
                            'ext': '.annot'
                        },
                        'folderId': {
                            'name': 'Image'
                        }
                    },
                    'disabled': True
                },
                'nuclei_annotation_format',
                'min_nucleus_area',
                'ignore_border_nuclei',
                'ImageInversionForm'
            ]
        }
    ]
)


def main():
    # Creates a multi-page application with a different set of slides on each page

    base_dir = 'C:\\Users\\samuelborder\\Desktop\\HIVE_Stuff\\FUSION\\Test Upload\\'
    local_slide_list = [
        base_dir+'XY01_IU-21-015F_001.svs',
        base_dir+'XY01_IU-21-015F.svs',
        base_dir+'new Visium\\V12U21-010_XY02_21-0069.tif',
        base_dir+'KPMP_Atlas_V2\\V10S14-085_XY01_20-0038\\V10S14-085_XY01_20-0038_lowres_image.tiff'
    ]

    # This can be set to False or removed, testing the upload overlaid annotations
    testing_upload = False
    local_annotations_list = [
        base_dir+'XY01_IU-21-015F_001.xml' if not testing_upload else None,
        None,
        base_dir+'new Visium\\V12U21-010_XY02_21-0069.h5ad',
        None
    ]

    dsa_handler = DSAHandler(
        girderApiUrl = 'http://ec2-3-230-122-132.compute-1.amazonaws.com:8080/api/v1'
    )

    dsa_items_list = [
        '64f545082d82d04be3e39ee1',
        '64f54be52d82d04be3e39f65',
        '64d3c6f3287cfdce1e9c4d88'
    ]

    dsa_tileservers = [dsa_handler.get_tile_server(i) for i in dsa_items_list]
    dsa_dataset_builder = dsa_handler.create_dataset_builder()
    dsa_login = dsa_handler.create_login_component()
    dsa_plugin_progress = dsa_handler.create_plugin_progress()
    dsa_uploader = dsa_handler.create_uploader(
        upload_types = [
            simple_upload_type,
            sequence_upload_type
        ]
    )

    vis_sess = Visualization(
        local_slides = local_slide_list,
        local_annotations = local_annotations_list,
        tileservers = dsa_tileservers,
        linkage = 'page',
        components = {
             "Visualization": [
                [
                    SlideMap(),
                    [
                        OverlayOptions(),
                        PropertyViewer(),
                        PropertyPlotter(),
                        HRAViewer(),
                        FeatureAnnotation(
                            storage_path = os.getcwd()+'\\tests\\Test_Annotations\\',
                            annotations_format = 'rgb'
                        ),
                        BulkLabels()
                    ]
                ]   
            ],
            "MultiFrame Visualization": [
                [
                    MultiFrameSlideMap(),
                    ChannelMixer()
                ]
            ],
            "Dataset Builder": [
                dsa_dataset_builder
            ],
            'Dataset Uploader': [
                dsa_uploader
            ]
        },
        header = [
            dsa_login,
            dsa_plugin_progress
        ],
        app_options={
            'port': 8050,
        }
    )

    vis_sess.start()

if __name__=='__main__':
    main()





"""Testing formatting intersecting masks
"""

import os
import sys
import json
sys.path.append('./src/')

from fusion_tools.utils.images import format_intersecting_masks


def main():

    path_to_test_all_annotations = 'C:\\Users\\samuelborder\\Downloads\\fusion-tools-current-layers.json'
    path_to_manual_roi = 'C:\\Users\\samuelborder\\Downloads\\Manual ROI 1.json'

    with open(path_to_test_all_annotations,'r') as f:
        all_anns = json.load(f)
        f.close()

    with open(path_to_manual_roi,'r') as f:
        manual_ann = json.load(f)
        f.close()

    mask_format = 'one-hot-labels'
    intersecting_masks = format_intersecting_masks(
        manual_ann,
        all_anns,
        mask_format
    )
    
    import matplotlib.pyplot as plt

    if not mask_format=='rgb':
        for m in intersecting_masks:
            for i in range(m.shape[-1]):
                plt.imshow(m[:,:,i])
                plt.show()
    else:
        for m in intersecting_masks:
            plt.imshow(m)
            plt.show()

if __name__=='__main__':
    main()


"""

Extracting nested properties for assembling tree_antd selectors.

Specifications:
    - Should work for arbitrary levels of nesting for each property
    - Should only include unique items

"""
import numpy as np

test_list = ['prop1','prop2','prop3', 'prop2',
             'prop4 --> sub_prop1', 'prop4 --> sub_prop2', 'prop4 --> sub_prop3', 'prop4 --> sub_prop3',
             'prop5 --> sub_prop1 --> sub_sub_prop1', 'prop5 --> sub_prop1 --> sub_sub_prop2', 'prop5 --> sub_prop2']

def generate_property_dict(all_property_list, title: str = 'Features', ignore_list: list = []):

    all_properties = {
        'title': title,
        'key': '0',
        'children': []
    }

    def add_prop_level(level_children, prop, index_list):
        new_keys = {}
        if len(level_children)==0:
            if not prop[0] in ignore_list:
                new_key = f'{"-".join(index_list)}-0'
                p_dict = {
                    'title': prop[0],
                    'key': new_key,
                    'children': []
                }
                l_dict = p_dict['children']
                new_keys[new_key] = prop[0]
                for p_idx,p in enumerate(prop[1:]):
                    if not p in ignore_list:
                        new_key = f'{"-".join(index_list+["0"]*(p_idx+2))}'
                        l_dict.append({
                            'title': p,
                            'key': new_key,
                            'children': []
                        })
                        l_dict = l_dict[0]['children']
                        new_keys[new_key] = ' --> '.join(prop[:p_idx+2])

                level_children.append(p_dict)
        else:
            for p_idx,p in enumerate(prop):
                if not p in ignore_list:
                    if any([p==i['title'] for i in level_children]):
                        title_idx = [i['title'] for i in level_children].index(p)
                        level_children = level_children[title_idx]['children']
                        index_list.append(str(title_idx))
                    else:
                        new_key = f'{"-".join(index_list)}-{len(level_children)}'
                        level_children.append({
                            'title': p,
                            'key': new_key,
                            'children': []
                        })
                        level_children = level_children[-1]['children']
                        index_list.append("0")
                        new_keys[new_key] = ' --> '.join(prop[:p_idx+1])
        
        return new_keys
    
    list_levels = [i.split(' --> ') if '-->' in i else [i] for i in all_property_list]
    unique_levels = list(set([len(i) for i in list_levels]))
    sorted_level_idxes = np.argsort(unique_levels)[::-1]
    property_keys = {}
    for s in sorted_level_idxes:
        depth_count = unique_levels[s]
        props_with_level = [i for i in list_levels if len(i)==depth_count]
        for p in props_with_level:
            feature_children = all_properties['children']
            property_keys = property_keys | add_prop_level(feature_children,p,['0'])
    
    return all_properties, property_keys

def main():

    ignore_list = ['prop3','sub_prop3']
    all_properties, property_keys = generate_property_dict(test_list,ignore_list = ignore_list)
    import json
    print(json.dumps(all_properties,indent=4))
    print(property_keys)




if __name__=='__main__':
    main()




"""Testing out LargeSlideMap components
"""

import os
import sys
sys.path.append('./src/')
import json
import requests

from fusion_tools.handler.dsa_handler import DSAHandler
from fusion_tools.visualization import Visualization
from fusion_tools.components.maps import LargeSlideMap, SlideMap
from fusion_tools.components import OverlayOptions



def main():

    # Grabbing first item from demo DSA instance
    base_url = 'http://ec2-3-230-122-132.compute-1.amazonaws.com:8080/api/v1'
    item_id = '64f545302d82d04be3e39eec'

    # Starting the DSAHandler to grab information:
    dsa_handler = DSAHandler(
        girderApiUrl = base_url
    )

    dsa_tileserver = [
        dsa_handler.get_tile_server(item_id)
    ]
    
    base_dir = 'C:\\Users\\samuelborder\\Desktop\\HIVE_Stuff\\FUSION\\Test Upload\\'
    local_slide_list = [
        base_dir+'XY01_IU-21-015F_001.svs'
    ]
    local_annotations_list = [
        base_dir+'XY01_IU-21-015F_001.xml'
    ]

    vis_session = Visualization(
        local_slides = local_slide_list,
        local_annotations = local_annotations_list,
        tileservers = dsa_tileserver,
        components = [
            [
                LargeSlideMap(
                    min_zoom = 4
                ),
                [
                    OverlayOptions()
                ]
            ]
        ]
    )

    vis_session.start()




















if __name__=='__main__':
    main()



"""Selective data grabber test
"""

import os
import sys
sys.path.append('./src/')
import json
import requests

from fusion_tools.visualization import Visualization
from fusion_tools.components import SlideMap, OverlayOptions, PropertyPlotter
from fusion_tools import Tool
from fusion_tools.visualization.vis_utils import get_pattern_matching_value

# Dash imports
import dash
dash._dash_renderer._set_react_version('18.2.0')
import dash_leaflet as dl
import dash_leaflet.express as dlx
from dash import dcc, callback, ctx, ALL, MATCH, exceptions, Patch, no_update, dash_table
from dash.dash_table.Format import Format, Scheme
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
import dash_treeview_antd as dta
from dash_extensions.enrich import DashBlueprint, html, Input, Output, State, PrefixIdTransform, MultiplexerTransform
from dash_extensions.javascript import Namespace, arrow_function


class DataExtractor(Tool):
    def __init__(self):

        super().__init__()

        self.exportable_session_data = {
            'Slide Metadata':{
                'description': 'This is any information about the current slide including labels, preparation details, and image tile information.'
            },
            'Visualization Session':{
                'description': 'This is a record of your current Visualization Session which can be uploaded in the Dataset Builder page to reload.'
            }
        }

        self.exportable_data = {
            'Properties':{
                'formats': ['CSV','XLSX','JSON'],
                'description': 'These are all per-structure properties and can include morphometrics, cell composition, channel intensity statistics, labels, etc.'
            },
            'Annotations':{
                'formats': ['Histomics (JSON)','GeoJSON','Aperio XML'],
                'description': 'These are the boundaries of selected structures. Different formats indicate the type of file generated and imported into another tool for visualization and analysis.'
            },
            'Images & Masks':{
                'formats': ['OME-TIFF'],
                'description': 'This will be a zip file containing combined images and masks for selected structures.'
            },
            'Images': {
                'formats': ['OME-TIFF','TIFF','PNG','JPG'],
                'description': 'This will be a zip file containing images of selected structures. Note: If this is a multi-frame image and OME-TIFF is not selected, images will be rendered in RGB according to the current colors in the map.'
            },
            'Masks': {
                'formats': ['OME-TIFF','TIFF','PNG','JPG'],
                'description': 'This will be a zip file containing masks of selected structures. Note: If a Manual ROI is selected, masks will include all intersecting structures as a separate label for each.'
            }
        }

    def __str__(self):
        return 'Data Extractor'

    def load(self, component_prefix:int):

        self.component_prefix = component_prefix

        self.title = 'Data Extractor'
        self.blueprint = DashBlueprint(
            transforms = [
                PrefixIdTransform(prefix = f'{self.component_prefix}',escape = lambda input_id: self.prefix_escape(input_id)),
                MultiplexerTransform()
            ]
        )

        self.get_callbacks()

    def get_scale_factors(self, image_metadata: dict):
        """Function used to initialize scaling factors applied to GeoJSON annotations to project annotations into the SlideMap CRS (coordinate reference system)

        :return: x and y (horizontal and vertical) scale factors applied to each coordinate in incoming annotations
        :rtype: float
        """

        base_dims = [
            image_metadata['sizeX']/(2**(image_metadata['levels']-1)),
            image_metadata['sizeY']/(2**(image_metadata['levels']-1))
        ]

        #x_scale = (base_dims[0]*(240/image_metadata['tileHeight'])) / image_metadata['sizeX']
        #y_scale = -((base_dims[1]*(240/image_metadata['tileHeight'])) / image_metadata['sizeY'])

        x_scale = base_dims[0] / image_metadata['sizeX']
        y_scale = -(base_dims[1]) / image_metadata['sizeY']


        return x_scale, y_scale

    def gen_layout(self,session_data:dict):

        # Set of components for importing markers
        # Set of components for adding markers
        # Set of components for selecting properties
        # Set of components for selecting additional data (image, geometry, bbox, slide name)
        # Set of components for downloading data

        layout = html.Div([
            dbc.Card(
                dbc.CardBody([
                    dbc.Row(
                        html.H3('Data Extractor')
                    ),
                    html.Hr(),
                    dbc.Row(
                        'Download select properties from indicated structures in the current slide.'
                    ),
                    html.Hr(),
                    html.Div([
                        dcc.Store(
                            id = {'type': 'data-extractor-store','index': 0},
                            storage_type = 'memory',
                            data = json.dumps({'selected_data': []})
                        ),
                        dbc.Modal(
                            id = {'type': 'data-extractor-download-modal','index': 0},
                            children = [],
                            is_open = [],
                            size = 'xl'
                        ),
                        dcc.Interval(
                            id = {'type': 'data-extractor-download-interval','index': 0},
                            disabled = True,
                            interval = 1000,
                            n_intervals = 0,
                            max_intervals=-1
                        )
                    ]),
                    dbc.Row([
                        dbc.Row([
                            dbc.Col(
                                html.H5('Download Session Data:'),
                                md = 5
                            ),
                            dbc.Col(
                                dcc.Dropdown(
                                    options = list(self.exportable_session_data.keys()),
                                    value = [],
                                    multi = False,
                                    placeholder = 'Select an option',
                                    id = {'type': 'data-extractor-session-data-drop','index': 0}
                                ),
                                md = 7
                            )
                        ]),
                        dbc.Row(
                            html.Div(
                                id = {'type': 'data-extractor-session-data-description','index': 0},
                                children = []
                            )
                        ),
                        dbc.Row(
                            dbc.Button(
                                id = {'type': 'data-extractor-download-session-data','index': 0},
                                n_clicks = 0,
                                className = 'd-grid col-12 mx-auto',
                                color = 'primary',
                                disabled = True
                            )
                        )
                    ],style = {'marginBottom':'10px'}),
                    html.Hr(),
                    dbc.Row([
                        dbc.Col(
                            html.H5('Select which structures to extract data from.'),
                            md = 9
                        ),
                        dbc.Col([
                            html.A(
                                html.I(
                                    className = 'fa-solid fa-rotate fa-xl',
                                    n_clicks = 0,
                                    id = {'type': 'data-extractor-refresh-icon','index': 0}
                                )
                            ),
                            dbc.Tooltip(
                                target = {'type': 'data-extractor-refresh-icon','index': 0},
                                children = 'Click to refresh available structures'
                            )
                        ],md = 3)
                    ],justify='left'),
                    dbc.Row([
                        dbc.Col([
                            dbc.Label('Current Structures:',html_for={'type':'data-extractor-current-structures-drop','index': 0})
                        ],md=3),
                        dbc.Col([
                            dcc.Dropdown(
                                options = [],
                                value = [],
                                multi = True,
                                placeholder = 'Structures in Slide',
                                id = {'type': 'data-extractor-current-structures-drop','index': 0}
                            )
                        ],md=9)
                    ]),
                    html.Hr(),
                    dbc.Row([
                        dbc.Col([
                            html.H5('Select what type of data you want to extract')
                        ])
                    ]),
                    dbc.Row([
                        dbc.Col(
                            dbc.Label('Available Data:',html_for={'type': 'data-extractor-available-data-drop','index': 0}),
                            md = 3
                        ),
                        dbc.Col(
                            dcc.Dropdown(
                                options = [],
                                value = [],
                                multi = True,
                                placeholder='Data',
                                id = {'type': 'data-extractor-available-data-drop','index':0}
                            ),
                            md = 9
                        )
                    ]),
                    html.Hr(),
                    html.Div(
                        id = {'type':'data-extractor-selected-data-parent-div','index': 0},
                        children = [
                            'Selected data descriptions will appear here'
                        ],
                        style = {'maxHeight': '100vh','overflow': 'scroll'}
                    ),
                    dbc.Row([
                        dcc.Loading([
                            dbc.Button(
                                className = 'd-grid col-12 mx-auto',
                                disabled = True,
                                color = 'primary',
                                id = {'type': 'data-extractor-download-button','index': 0}
                            ),
                            dcc.Download(
                                id = {'type': 'data-extractor-download','index': 0}
                            )
                        ])
                    ],style = {'marginTop':'10px'})
                ])
            )
        ])

        self.blueprint.layout = layout

    def get_callbacks(self):
        
        # Callback for updating current slide
        self.blueprint.callback(
            [
                Input({'type': 'map-annotations-store','index': ALL},'data')
            ],
            [
                Output({'type': 'data-extractor-current-structures-drop','index': ALL},'options'),
                Output({'type': 'data-extractor-available-data-drop','index': ALL},'options'),
                Output({'type': 'data-extractor-selected-data-parent','index': ALL},'children'),
                Output({'type': 'data-extractor-download-button','index': ALL},'disabled'),
                Output({'type':'data-extractor-store','index': ALL},'data')
            ],
            prevent_initial_call = True
        )(self.update_slide)

        # Callback for the refresh button being pressed
        self.blueprint.callback(
            [
                Input({'type': 'data-extractor-refresh-icon','index': ALL},'n_clicks')
            ],
            [
                Output({'type': 'data-extractor-current-structures-drop','index': ALL},'options'),
                Output({'type': 'data-extractor-available-data-drop','index': ALL},'options'),
                Output({'type': 'data-extractor-selected-data-parent','index': ALL},'children'),
                Output({'type': 'data-extractor-download-button','index': ALL},'disabled'),
                Output({'type': 'data-extractor-store','index': ALL},'data')
            ],
            [
                State({'type': 'feature-overlay','index': ALL},'name'),
                State({'type': 'map-marker-div','index': ALL},'children')
            ],
            prevent_initial_call = True
        )(self.refresh_data)

        # Callback for selecting a session data item
        self.blueprint.callback(
            [
                Input({'type':'data-extractor-session-data-drop','index': ALL},'value')
            ],
            [
                Output({'type': 'data-extractor-session-data-description','index': ALL},'children'),
                Output({'type': 'data-extractor-download-session-data','index': ALL},'disabled')
            ]
        )(self.update_session_data_description)

        # Callback for downloading session data item
        self.blueprint.callback(
            [
                Input({'type': 'data-extractor-download-session-data','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'data-extractor-session-data-drop','index': ALL},'value'),
                State({'type': 'map-slide-information','index':ALL},'data'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'data-extractor-download','index': ALL},'data')
            ]
        )(self.download_session_data)

        # Callback for updating selected data information and enabling download button
        self.blueprint.callback(
            [
                Input({'type': 'data-extractor-available-data-drop','index': ALL},'value')
            ],
            [
                State({'type': 'data-extractor-store','index': ALL},'data')
            ],
            [
                Output({'type': 'data-extractor-selected-data-parent','index': ALL},'children'),
                Output({'type': 'data-extractor-download-button','index': ALL},'disabled'),
                Output({'type': 'data-extractor-store','index': ALL},'data')
            ],
            prevent_initial_call = True
        )(self.update_data_info)

        # Callback for downloading selected data and clearing selections
        self.blueprint.callback(
            [
                Input({'type': 'data-extractor-download-button','index': ALL},'n_clicks')
            ],
            [
                Output({'type': 'data-extractor-download-interval','index': ALL},'disabled'),
                Output({'type': 'data-extractor-download','index': ALL},'data'),
                Output({'type':'data-extractor-current-structures-drop','index': ALL},'value'),
                Output({'type': 'data-extractor-available-data-drop','index': ALL},'value'),
                Output({'type': 'data-extractor-selected-data-parent','index': ALL},'children'),
                Output({'type': 'data-extractor-download-button','index': ALL},'disabled'),
                Output({'type': 'data-extractor-store','index': ALL},'data')
            ],
            [
                State({'type': 'data-extractor-current-structures-drop','index': ALL},'value'),
                State({'type': 'data-extractor-available-data-drop','index': ALL},'value'),
                State({'type': 'data-extractor-selected-data-format','index': ALL},'value'),
                State({'type': 'map-annotations-store','index':ALL},'data'),
                State({'type': 'map-marker-div','index': ALL},'children'),
                State({'type': 'map-slide-information','index': ALL},'data'),
                State({'type': 'base-layer','index': ALL},'checked'),
                State({'type': 'tile-layer','index': ALL},'url')
            ],
            prevent_initial_call = True
        )(self.start_download_data)

        self.blueprint.callback(
            [
                Input({'type': 'data-extractor-download-interval','index': ALL},'n_intervals')
            ],
            [
                Output({'type':'data-extractor-download-interval','index': ALL},'disabled'),
                Output({'type': 'data-extractor-download-modal','index': ALL},'is_open'),
                Output({'type': 'data-extractor-download-modal','index':ALL},'children'),
                Output({'type': 'data-extractor-download','index': ALL},'data')
            ],
            prevent_initial_call = True
        )(self.update_download_data)
    
    def update_slide(self, new_annotations:list):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        new_annotations = json.loads(get_pattern_matching_value(new_annotations))
        new_structure_names = [i['properties']['name'] for i in new_annotations if 'properties' in i]

        available_structures_drop = [
            {'label': i, 'value': i, 'disabled': False}
            for i in new_structure_names
        ]
        available_structures_drop += [{'label': 'Marked Structures','value': 'Marked Structures','disabled': True}]

        available_data_drop = [
            {'label': i, 'value': i, 'disabled': False}
            for i in self.exportable_data
        ]

        button_disabled = True
        new_data_extractor_store = json.dumps({'selected_data': []})

        return [available_structures_drop], [available_data_drop], ['Selected data descriptions will appear here'], [button_disabled], [new_data_extractor_store]

    def refresh_data(self, clicked, overlay_names, marker_div_children):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        print(overlay_names)
        print(marker_div_children)
        available_structures_drop = [
            {'label': i, 'value': i, 'disabled': False}
            for i in get_pattern_matching_value(overlay_names)
        ]
        available_structures_drop += [{'label': 'Marked Structures','value': 'Marked Structures','disabled': True if len(marker_div_children)==0 else False}]
        
        available_data_drop = [
            {'label': i, 'value': i, 'disabled': False}
            for i in self.exportable_data
        ]

        button_disabled = True

        new_data_extractor_store = json.dumps({'selected_data': []})


        return [available_structures_drop], [available_data_drop], ['Selected data descriptions will appear here'], [button_disabled],[new_data_extractor_store]

    def update_session_data_description(self, session_data_selection):
        
        if not any([i['value'] for i in ctx.triggered]):
            return ['Select a type of session data to see a description']
        
        session_data_selection = get_pattern_matching_value(session_data_selection)

        return [self.exportable_session_data[session_data_selection]['description']]

    def download_session_data(self, clicked, session_data_selection, current_slide_info, session_data):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        session_data_selection = get_pattern_matching_value(session_data_selection)
        current_slide_info = json.loads(get_pattern_matching_value(current_slide_info))
        session_data = json.loads(session_data)

        if session_data_selection == 'Slide Metadata':
            slide_tile_url = current_slide_info['tiles_url']
            current_slide_tile_urls = [i['tiles_url'] for i in session_data['current']]
            slide_idx = current_slide_tile_urls.index(slide_tile_url)

            slide_metadata = requests.get(session_data['current'][slide_idx]['metadata_url'])

            download_content = {'content': json.dumps(slide_metadata,indent=4),'filename': 'slide_metadata.json'}
        elif session_data_selection == 'Visualization Session':
            download_content = {'content': json.dumps(session_data,indent=4),'filename': 'fusion_visualization_session.json'}

        return [download_content]

    def update_data_info(self, selected_data, data_extract_store):

        if not any([i['value'] for i in ctx.triggered]):
            return ['Selected data descriptions will appear here']
        
        print(selected_data)
        selected_data = get_pattern_matching_value(selected_data)
        current_selected_data = json.loads(get_pattern_matching_value(data_extract_store))['selected_data']
        print(current_selected_data)

        def make_new_info(data_type):
            data_type_index = list(self.exportable_data.keys()).index(data_type)
            return_card = dbc.Card([
                dbc.CardHeader(data_type),
                dbc.CardBody([
                    dbc.Row([
                        dbc.Col(
                            dbc.Label('Download Format: '),
                            md = 3
                        ),
                        dbc.Col(
                            dcc.Dropdown(
                                options = self.exportable_data[data_type]['formats'],
                                value = self.exportable_data[data_type]['formats'][0],
                                multi = False,
                                placeholder = 'Select a format',
                                id = {'type': f'{self.component_prefix}-data-extractor-selected-data-format','index': data_type_index}
                            ),
                            md = 9
                        )
                    ])
                ])
            ])

            return return_card

        # Which ones are in selected_data that aren't in current_selected_data
        new_selected = list(set(selected_data) - set(current_selected_data))
        info_return = Patch()
        if len(new_selected)>0:
            # Appending to the partial property update
            info_return.append(make_new_info(new_selected))

        else:
            # Removing some de-selected values
            removed_type = list(set(current_selected_data)-set(selected_data))

            del info_return[current_selected_data.index(removed_type)]
        
        return [info_return]

    def start_download_data(self, clicked, selected_structures, selected_data, selected_data_formats, slide_annotations, slide_markers, slide_info, base_layer_checked, tile_layer_urls):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        interval_disabled = False
        new_values = []
        button_disabled = True
        selected_structures = get_pattern_matching_value(selected_structures)
        selected_data = get_pattern_matching_value(selected_data)
        slide_annotations = json.loads(get_pattern_matching_value(slide_annotations))


        print(selected_structures)
        print(selected_data)
        print(selected_data_formats)

        # Starting separate threads for downloads
        for s_idx, struct in enumerate(selected_structures):
            for d_idx, data in enumerate(selected_data):
                print(f'Starting thread for {struct} {data}')


        return [interval_disabled], [new_values], [new_values], ['Selected data descriptions will appear here'], [button_disabled]

    def update_download_data(self, new_interval):
        
        new_interval = get_pattern_matching_value(new_interval)

        if new_interval<5: 
            interval_disabled = False
            modal_open = True
            modal_children = [f'Interval count: {new_interval}/5']
            download_data = no_update
        else:
            interval_disabled = True
            modal_open = False
            modal_children = []
            download_data = no_update
        
        return [interval_disabled], [modal_open], [modal_children], [download_data]
        


def main():
    pass


if __name__=='__main__':
    main()





"""

Testing remote tile server with annotations


"""

import os
import sys
import threading
sys.path.append('./src/')
from fusion_tools import Visualization, DSAHandler
from fusion_tools.tileserver import DSATileServer
from fusion_tools.components import SlideMap


def main():

    # Grabbing first item from demo DSA instance
    base_url = 'https://demo.kitware.com/histomicstk/api/v1'
    item_id = '5bbdeed1e629140048d01bcb'

    # Starting visualization session
    tile_server = DSATileServer(
        api_url = base_url,
        item_id = item_id
    )

    # Starting the DSAHandler to grab information:
    dsa_handler = DSAHandler(
        girderApiUrl = base_url
    )

    # Checking how many annotations this item has:
    #print('This item has the following annotations: ')
    #print(dsa_handler.query_annotation_count(item=item_id).to_dict('records'))

    annotations = dsa_handler.get_annotations(
        item = item_id
    )

    vis_session = Visualization(
        components = [
            [
                SlideMap(
                    tile_server = tile_server,
                    annotations = annotations
                )
            ]
        ]
    )

    vis_session.start()


if __name__=='__main__':
    main()

